# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import ovn_stretched_l2_bgp_driver
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests.unit import fakes
from ovn_bgp_agent.utils import linux_net

import ipaddress

CONF = cfg.CONF


class TestHashedRoute(test_base.TestCase):
    def setUp(self):
        super(TestHashedRoute, self).setUp()
        self.table = set()
        self.route = ovn_stretched_l2_bgp_driver.HashedRoute(
            "192.168.0.0", 24, "192.168.1.1")
        self.invalid_route = ovn_stretched_l2_bgp_driver.HashedRoute(
            "192.168.0.0", 24, "192.168.1.2")
        self.table.add(self.route)

    def test_lookup(self):
        self.assertIn(self.route, self.table)
        self.assertNotIn(self.invalid_route, self.table)

    def test_delete(self):
        self.table.remove(self.route)
        self.assertEqual(0, len(self.table))


class TestOVNBGPStretchedL2Driver(test_base.TestCase):
    def setUp(self):
        super(TestOVNBGPStretchedL2Driver, self).setUp()
        CONF.set_override(
            "address_scopes",
            "11111111-1111-1111-1111-11111111,22222222-2222-2222-2222-22222222",  # NOQA E501
        )
        self.bgp_driver = ovn_stretched_l2_bgp_driver.OVNBGPStretchedL2Driver()
        self.bgp_driver._post_fork_event = mock.Mock()
        self.bgp_driver.sb_idl = mock.Mock()
        self.sb_idl = self.bgp_driver.sb_idl
        self.bgp_driver.chassis = "fake-chassis"
        # self.bgp_driver.ovn_routing_tables = {self.bridge: 'fake-table'}
        # self.bgp_driver.ovn_bridge_mappings = {'fake-network': self.bridge}

        self.mock_sbdb = mock.patch.object(ovn, "OvnSbIdl").start()
        self.mock_ovs_idl = mock.patch.object(ovs, "OvsIdl").start()
        self.ipv4 = "192.168.1.17"
        self.ipv6 = "2002::1234:abcd:ffff:c0a8:101"
        self.fip = "172.24.4.33"
        self.mac = "aa:bb:cc:dd:ee:ff"
        self.bgp_driver.ovs_idl = self.mock_ovs_idl

        self.test_route_ipv4 = ovn_stretched_l2_bgp_driver.HashedRoute(
            network="192.168.1.0",
            prefix_len=24,
            dst="10.0.0.1",
        )

        self.test_route_ipv6 = ovn_stretched_l2_bgp_driver.HashedRoute(
            network="fdcc:8cf2:d40c:2::",
            prefix_len=64,
            dst="fd51:f4b3:872:eda::1",
        )

        self.addr_scopev4 = "11111111-1111-1111-1111-11111111"
        self.addr_scopev6 = "22222222-2222-2222-2222-22222222"
        self.addr_scope = {
            constants.IP_VERSION_4: self.addr_scopev4,
            constants.IP_VERSION_6: self.addr_scopev6,
        }

        self.addr_scope_external_ids = {
            "neutron:subnet_pool_addr_scope4": self.addr_scopev4,
            "neutron:subnet_pool_addr_scope6": self.addr_scopev6,
        }

        self.cr_lrp0 = mock.Mock()
        self.cr_lrp0.mac = [
            "ff:ff:ff:ff:ff:00 10.0.0.1/24 fd51:f4b3:872:eda::1/64"
        ]
        self.cr_lrp0.datapath = "fake-router-dp"
        self.cr_lrp0.type = constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE
        self.cr_lrp0.logical_port = "cr-lrp-fake-port"

        self.lp0 = mock.Mock()
        self.lp0.external_ids = self.addr_scope_external_ids

        self.router_port = fakes.create_object(
            {
                "name": "fake-router-port",
                "mac": [
                    "ff:ff:ff:ff:ff:01 192.168.1.1/24 fdcc:8cf2:d40c:2::1/64"
                ],
                "logical_port": "lrp-fake-logical-port",
            }
        )

        self.fake_patch_port = fakes.create_object(
            {
                "name": "fake-patch-port",
                "mac": [
                    "ff:ff:ff:ff:ff:01 192.168.1.1/24 fdcc:8cf2:d40c:2::1/64"
                ],
                "external_ids": self.addr_scope_external_ids,
                "logical_port": "fake-port",
            }
        )

        # Mock pyroute2.NDB context manager object
        self.mock_ndb = mock.patch.object(linux_net.pyroute2, "NDB").start()
        self.fake_ndb = self.mock_ndb().__enter__()

    @mock.patch.object(linux_net, "ensure_vrf")
    @mock.patch.object(linux_net, "ensure_ovn_device")
    @mock.patch.object(linux_net, "delete_routes_from_table")
    @mock.patch.object(frr, "vrf_leak")
    def test_start(self, mock_vrf, mock_delete_routes, mock_ensure_ovn_device,
                   *args):
        CONF.set_override("clear_vrf_routes_on_startup", True)

        mock_redistribute = mock.patch.object(
            frr, "set_default_redistribute"
        ).start()

        self.bgp_driver.start()

        mock_redistribute.assert_called_with(['kernel'])
        mock_vrf.assert_called_once_with(
            CONF.bgp_vrf,
            CONF.bgp_AS,
            CONF.bgp_router_id,
            template=frr.LEAK_VRF_TEMPLATE,
        )
        # Assert connections were started
        self.mock_ovs_idl().start.assert_called_once_with(
            CONF.ovsdb_connection
        )
        self.mock_sbdb().start.assert_called_once_with()
        mock_delete_routes.assert_called_once_with(CONF.bgp_vrf_table_id)
        mock_ensure_ovn_device.assert_called_once_with(
            CONF.bgp_nic, CONF.bgp_vrf)

    @mock.patch.object(linux_net, "ensure_vrf")
    @mock.patch.object(linux_net, "ensure_ovn_device")
    @mock.patch.object(linux_net, "delete_routes_from_table")
    @mock.patch.object(frr, "vrf_leak")
    def test_start_clear_routes(
        self, mock_vrf, mock_delete_routes, mock_ensure_ovn_device, *args):
        CONF.set_override("clear_vrf_routes_on_startup", False)

        mock_redistribute = mock.patch.object(
            frr, "set_default_redistribute"
        ).start()

        self.bgp_driver.start()

        mock_redistribute.assert_called_with(['kernel'])
        mock_vrf.assert_called_once_with(
            CONF.bgp_vrf,
            CONF.bgp_AS,
            CONF.bgp_router_id,
            template=frr.LEAK_VRF_TEMPLATE,
        )
        # Assert connections were started
        self.mock_ovs_idl().start.assert_called_once_with(
            CONF.ovsdb_connection
        )
        self.mock_sbdb().start.assert_called_once_with()
        mock_delete_routes.assert_not_called()
        mock_ensure_ovn_device.assert_called_once_with(
            CONF.bgp_nic, CONF.bgp_vrf)

    @mock.patch.object(linux_net, "add_ip_route")
    def test__add_route(self, mock_add_route):
        for test_route in [self.test_route_ipv4, self.test_route_ipv6]:
            self.bgp_driver._add_route(
                test_route.network,
                test_route.prefix_len,
                test_route.dst,
            )

            mock_add_route.assert_called_with(
                mock.ANY,
                test_route.network,
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=test_route.prefix_len,
                via=test_route.dst,
            )

            self.assertIn(test_route, self.bgp_driver.vrf_routes)

    @mock.patch.object(linux_net, "del_ip_route")
    def test__del_route(self, mock_del_route):
        self.bgp_driver.vrf_routes.add(self.test_route_ipv4)
        self.bgp_driver.vrf_routes.add(self.test_route_ipv6)
        for test_route in [self.test_route_ipv4, self.test_route_ipv6]:
            self.bgp_driver._del_route(
                test_route.network,
                test_route.prefix_len,
                test_route.dst,
            )

            mock_del_route.assert_called_with(
                mock.ANY,
                test_route.network,
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=test_route.prefix_len,
                via=test_route.dst,
            )

            self.assertNotIn(test_route, self.bgp_driver.vrf_routes)

    def test__address_scope_allowed(self):
        test_scope2 = {
            constants.IP_VERSION_4: self.addr_scopev4,
            constants.IP_VERSION_6: "33333333-3333-3333-3333-33333333",
        }

        self.assertTrue(
            self.bgp_driver._address_scope_allowed(
                self.addr_scope,
                test_scope2,
                constants.IP_VERSION_4,
            )
        )

        self.assertFalse(
            self.bgp_driver._address_scope_allowed(
                self.addr_scope,
                test_scope2,
                constants.IP_VERSION_6,
            )
        )

    def test__address_scope_not_allowed_scope(self):
        test_scope1 = {
            constants.IP_VERSION_4: "33333333-3333-3333-3333-33333333",
            constants.IP_VERSION_6: "33333333-3333-3333-3333-33333333",
        }

        test_scope2 = {
            constants.IP_VERSION_4: "33333333-3333-3333-3333-33333333",
            constants.IP_VERSION_6: "33333333-3333-3333-3333-33333333",
        }

        self.assertFalse(
            self.bgp_driver._address_scope_allowed(
                test_scope1,
                test_scope2,
                constants.IP_VERSION_4,
            )
        )

        self.assertFalse(
            self.bgp_driver._address_scope_allowed(
                test_scope1,
                test_scope2,
                constants.IP_VERSION_6,
            )
        )

    def test__address_scope_allowed_no_scope(self):
        self.bgp_driver.allowed_address_scopes = set()
        test_scope2 = {
            constants.IP_VERSION_4: None,
            constants.IP_VERSION_6: None,
        }

        self.assertTrue(
            self.bgp_driver._address_scope_allowed(
                self.addr_scope,
                test_scope2,
                constants.IP_VERSION_4,
            )
        )

        self.assertTrue(
            self.bgp_driver._address_scope_allowed(
                self.addr_scope,
                test_scope2,
                constants.IP_VERSION_6,
            )
        )

    def test__address_scope_allowed_no_match(self):
        test_scope2 = {
            constants.IP_VERSION_4: None,
            constants.IP_VERSION_6: "44444444-4444-4444-4444-44444444",
        }

        self.assertFalse(
            self.bgp_driver._address_scope_allowed(
                self.addr_scope,
                test_scope2,
                constants.IP_VERSION_4,
            )
        )

        self.assertFalse(
            self.bgp_driver._address_scope_allowed(
                self.addr_scope,
                test_scope2,
                constants.IP_VERSION_6,
            )
        )

    def test_expose_subnet(self):
        mock__ensure_network_exposed = mock.patch.object(
            self.bgp_driver, "_ensure_network_exposed"
        ).start()
        self.sb_idl.is_router_gateway_on_any_chassis.return_value = (
            self.cr_lrp0
        )
        row = mock.Mock()
        row.datapath = "fake-dp"

        self.bgp_driver.expose_subnet(None, row)

        self.sb_idl.is_router_gateway_on_any_chassis.assert_called_once_with(
            row.datapath
        )

        mock__ensure_network_exposed.assert_called_once_with(
            row, self.cr_lrp0.logical_port
        )

    def test_expose_subnet_no_gateway_port(self):
        mock__ensure_network_exposed = mock.patch.object(
            self.bgp_driver, "_ensure_network_exposed"
        ).start()
        self.sb_idl.is_router_gateway_on_any_chassis.return_value = None
        row = mock.Mock()
        row.datapath = "fake-dp"

        self.bgp_driver.expose_subnet(None, row)

        self.sb_idl.is_router_gateway_on_any_chassis.assert_called_once_with(
            row.datapath
        )

        mock__ensure_network_exposed.assert_not_called()

    def test_expose_subnet_no_datapath(self):
        mock__ensure_network_exposed = mock.patch.object(
            self.bgp_driver, "_ensure_network_exposed"
        ).start()
        row = mock.Mock()
        row.datapath = "fake-dp"
        self.sb_idl.is_router_gateway_on_any_chassis.side_effect = (
            agent_exc.DatapathNotFound(datapath=row.datapath))

        self.bgp_driver.expose_subnet(None, row)

        self.sb_idl.is_router_gateway_on_any_chassis.assert_called_once_with(
            row.datapath
        )

        mock__ensure_network_exposed.assert_not_called()

    def test_update_subnet(self):
        mock__update_network = mock.patch.object(
            self.bgp_driver, "_update_network"
        ).start()
        self.sb_idl.is_router_gateway_on_any_chassis.return_value = (
            self.cr_lrp0
        )
        old = mock.Mock()
        old.mac = ["ff:ff:ff:ff:ff:01 1.1.1.1/24 2.2.2.2/24"]

        row = mock.Mock()
        row.datapath = "fake-dp"
        row.mac = ["ff:ff:ff:ff:ff:01 2.2.2.2/24 3.3.3.3/24"]

        self.bgp_driver.update_subnet(old, row)

        self.sb_idl.is_router_gateway_on_any_chassis.assert_called_once_with(
            row.datapath
        )

        mock__update_network.assert_called_once_with(
            row, self.cr_lrp0.logical_port, ["3.3.3.3/24"], ["1.1.1.1/24"]
        )

    def test_update_subnet_no_datapath(self):
        mock__update_network = mock.patch.object(
            self.bgp_driver, "_update_network"
        ).start()
        self.sb_idl.is_router_gateway_on_any_chassis.return_value = (
            self.cr_lrp0
        )
        old = mock.Mock()
        old.mac = ["ff:ff:ff:ff:ff:01 1.1.1.1/24 2.2.2.2/24"]

        row = mock.Mock()
        row.datapath = "fake-dp"
        row.mac = ["ff:ff:ff:ff:ff:01 2.2.2.2/24 3.3.3.3/24"]

        self.sb_idl.is_router_gateway_on_any_chassis.side_effect = (
            agent_exc.DatapathNotFound(datapath=row.datapath))

        self.bgp_driver.update_subnet(old, row)

        self.sb_idl.is_router_gateway_on_any_chassis.assert_called_once_with(
            row.datapath
        )

        mock__update_network.assert_not_called()

    @mock.patch.object(linux_net, "get_exposed_routes_on_network")
    @mock.patch.object(linux_net, "del_ip_route")
    @mock.patch.object(linux_net, "add_ip_route")
    def test__update_network(
        self,
        mock_add_ip_route,
        mock_del_ip_route,
        mock_get_exposed_routes_on_network,
    ):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        gateway["lrp_ports"] = set()
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.router_lrp = mock.Mock()
        self.router_lrp.mac = [
            "ff:ff:ff:ff:ff:01 192.168.1.1/24 fdcc:8cf2:d40c:2::1/64"
        ]

        add_ips = ["192.168.1.1/24", "fdcc:8cf2:d40c:2::1/64"]
        delete_ips = ["192.168.0.1/24"]

        mock_get_exposed_routes_on_network.side_effect = (
            ["route-v4"],
            ["route-v6"],
        )

        self.sb_idl.get_port_by_name.return_value = self.fake_patch_port

        self.bgp_driver._update_network(
            self.router_port, "gateway_port", add_ips, delete_ips
        )

        self.sb_idl.get_port_by_name.assert_called_once_with(
            "fake-logical-port"
        )

        expected_calls = [
            mock.call(
                mock.ANY,
                "10.0.0.0",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=26,
                via=None,
            ),
            mock.call(
                mock.ANY,
                "192.168.1.0",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=24,
                via="10.0.0.10",
            ),
            mock.call(
                mock.ANY,
                "fd51:f4b3:872:eda::",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=64,
                via=None,
            ),
            mock.call(
                mock.ANY,
                "fdcc:8cf2:d40c:2::",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=64,
                via="fd51:f4b3:872:eda::10",
            ),
        ]

        mock_add_ip_route.assert_has_calls(expected_calls)
        mock_del_ip_route.assert_called_once_with(
            mock.ANY,
            "192.168.0.0",
            CONF.bgp_vrf_table_id,
            CONF.bgp_nic,
            vlan=None,
            mask=24,
            via="10.0.0.10",
        )

        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {
                self.router_port.logical_port: {
                    "cr_lrp": "gateway_port",
                    "subnets": {
                        "fdcc:8cf2:d40c:2::/64",
                        "192.168.1.0/24"
                    }
                }
            }
        )

    @mock.patch.object(linux_net, "get_exposed_routes_on_network")
    @mock.patch.object(linux_net, "del_ip_route")
    @mock.patch.object(linux_net, "add_ip_route")
    def test__update_network_no_gateway(
        self,
        mock_add_ip_route,
        mock_del_ip_route,
        mock_get_exposed_routes_on_network,
    ):
        self.bgp_driver.ovn_local_cr_lrps = {}

        self.router_lrp = mock.Mock()
        self.router_lrp.mac = [
            "ff:ff:ff:ff:ff:01 192.168.1.1/24 fdcc:8cf2:d40c:2::1/64"
        ]

        add_ips = ["192.168.1.1/24", "fdcc:8cf2:d40c:2::1/64"]
        delete_ips = ["192.168.0.1/24"]

        self.bgp_driver._update_network(
            self.router_port, "gateway_port", add_ips, delete_ips
        )

        mock_get_exposed_routes_on_network.assert_not_called()
        mock_del_ip_route.assert_not_called()
        mock_add_ip_route.assert_not_called()
        self.sb_idl.get_port_by_name.assert_not_called()

    @mock.patch.object(linux_net, "get_exposed_routes_on_network")
    @mock.patch.object(linux_net, "del_ip_route")
    @mock.patch.object(linux_net, "add_ip_route")
    def test__update_network_no_mac(
        self,
        mock_add_ip_route,
        mock_del_ip_route,
        mock_get_exposed_routes_on_network,
    ):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.router_port.mac = []

        add_ips = ["192.168.1.1/24", "fdcc:8cf2:d40c:2::1/64"]
        delete_ips = ["192.168.0.1/24"]

        self.bgp_driver._update_network(
            self.router_port, "gateway_port", add_ips, delete_ips
        )

        mock_get_exposed_routes_on_network.assert_not_called()
        mock_del_ip_route.assert_not_called()
        mock_add_ip_route.assert_not_called()
        self.sb_idl.get_port_by_name.assert_not_called()

    def test_withdraw_subnet(self):
        mock__withdraw_subnet = mock.patch.object(
            self.bgp_driver, "_withdraw_subnet"
        ).start()

        row = mock.Mock()
        row.datapath = "fake-dp"
        row.logical_port = "fake-lport"
        port_info = {
            "cr_lrp": self.cr_lrp0.logical_port,
            "subnets": {
                "fdcc:8cf2:d40c:2::/64",
                "192.168.1.0/24"
            }
        }

        self.bgp_driver.propagated_lrp_ports = {
            row.logical_port: port_info,
            "another_lrp_port": {}
        }
        self.bgp_driver.ovn_local_cr_lrps = {
            self.cr_lrp0.logical_port: {
                "lrp_ports": {row.logical_port, "another_lrp_port"}
            }
        }

        self.bgp_driver.withdraw_subnet(None, row)

        mock__withdraw_subnet.assert_called_once_with(
            port_info, self.cr_lrp0.logical_port
        )
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {
                "another_lrp_port": {}
            }
        )
        self.assertDictEqual(
            self.bgp_driver.ovn_local_cr_lrps,
            {
                self.cr_lrp0.logical_port: {
                    "lrp_ports": {"another_lrp_port"}
                }
            }
        )

    @mock.patch.object(linux_net, "add_ip_route")
    def test__ensure_network_exposed(self, mock_add_ip_route):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        gateway["lrp_ports"] = set()
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.router_lrp = mock.Mock()
        self.router_lrp.mac = [
            "ff:ff:ff:ff:ff:01 192.168.1.1/24 fdcc:8cf2:d40c:2::1/64"
        ]

        self.sb_idl.get_port_by_name.return_value = self.fake_patch_port

        self.bgp_driver._ensure_network_exposed(
            self.router_port, "gateway_port"
        )

        self.sb_idl.get_port_by_name.assert_called_once_with(
            "fake-logical-port"
        )

        expected_calls = [
            mock.call(
                mock.ANY,
                "10.0.0.0",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=26,
                via=None,
            ),
            mock.call(
                mock.ANY,
                "192.168.1.0",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=24,
                via="10.0.0.10",
            ),
            mock.call(
                mock.ANY,
                "fd51:f4b3:872:eda::",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=64,
                via=None,
            ),
            mock.call(
                mock.ANY,
                "fdcc:8cf2:d40c:2::",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=64,
                via="fd51:f4b3:872:eda::10",
            ),
        ]

        mock_add_ip_route.assert_has_calls(expected_calls)
        self.assertDictEqual(
            self.bgp_driver.ovn_local_cr_lrps,
            {
                'gateway_port': {
                    'address_scopes': {
                        4: '11111111-1111-1111-1111-11111111',
                        6: '22222222-2222-2222-2222-22222222'},
                    'ips': [
                        ipaddress.IPv4Interface('10.0.0.10/26'),
                        ipaddress.IPv6Interface('fd51:f4b3:872:eda::10/64')],
                    'lrp_ports': {'lrp-fake-logical-port'}
                }
            }
        )
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {
                "lrp-fake-logical-port": {
                    'cr_lrp': 'gateway_port',
                    'subnets': {'192.168.1.0/24', 'fdcc:8cf2:d40c:2::/64'}
                }
            }
        )

    @mock.patch.object(linux_net, "add_ip_route")
    def test__ensure_network_exposed_invalid_addr_scopes(
        self,
        mock_add_ip_route
    ):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]

        # Both of them are valid but none of them matches to the correct
        # IP version
        gateway["address_scopes"] = {
            constants.IP_VERSION_4: self.addr_scopev6,
            constants.IP_VERSION_6: self.addr_scopev4,
        }
        gateway["lrp_ports"] = set()
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.router_lrp = mock.Mock()
        self.router_lrp.mac = [
            "ff:ff:ff:ff:ff:01 192.168.1.1/24 fdcc:8cf2:d40c:2::1/64"
        ]

        self.sb_idl.get_port_by_name.return_value = self.fake_patch_port

        self.bgp_driver._ensure_network_exposed(
            self.router_port, "gateway_port"
        )

        self.sb_idl.get_port_by_name.assert_called_once_with(
            "fake-logical-port"
        )
        mock_add_ip_route.assert_not_called()
        self.assertDictEqual(
            self.bgp_driver.ovn_local_cr_lrps,
            {
                'gateway_port': {
                    'address_scopes': {
                        4: '22222222-2222-2222-2222-22222222',
                        6: '11111111-1111-1111-1111-11111111'},
                    'ips': [
                        ipaddress.IPv4Interface('10.0.0.10/26'),
                        ipaddress.IPv6Interface('fd51:f4b3:872:eda::10/64')],
                    'lrp_ports': set()
                }
            }
        )
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {}
        )

    @mock.patch.object(linux_net, "add_ip_route")
    def test__ensure_network_exposed_no_gateway(self, mock_add_ip_route):
        self.bgp_driver.ovn_local_cr_lrps = {}

        self.bgp_driver._ensure_network_exposed(
            self.router_port, "gateway_port"
        )

        self.sb_idl.get_port_by_name.assert_not_called()
        mock_add_ip_route.assert_not_called()
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {}
        )

    @mock.patch.object(linux_net, "add_ip_route")
    def test__ensure_network_exposed_duplicate_ip(self, mock_add_ip_route):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["192.168.1.1/24", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.bgp_driver._ensure_network_exposed(
            self.router_port, "gateway_port"
        )

        self.sb_idl.get_port_by_name.assert_not_called()
        mock_add_ip_route.assert_not_called()
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {}
        )

    @mock.patch.object(driver_utils, "get_addr_scopes")
    @mock.patch.object(linux_net, "add_ip_route")
    def test__ensure_network_exposed_port_not_existing(self,
                                                       mock_add_ip_route,
                                                       mock_addr_scopes):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.sb_idl.get_port_by_name.return_value = []
        self.bgp_driver._ensure_network_exposed(
            self.router_port, "gateway_port"
        )
        mock_addr_scopes.assert_not_called()
        mock_add_ip_route.assert_not_called()
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {}
        )

    @mock.patch.object(linux_net, "add_ip_route")
    def test__ensure_network_exposed_port_addr_scope_no_match(
        self,
        mock_add_ip_route
    ):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = {
            constants.IP_VERSION_4: "address_scope_v4",
            constants.IP_VERSION_6: "address_scope_v6",
        }
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}

        self.sb_idl.get_port_by_name.return_value = self.fake_patch_port
        self.bgp_driver._ensure_network_exposed(
            self.router_port, "gateway_port"
        )

        self.sb_idl.get_port_by_name.assert_called_once_with(
            "fake-logical-port"
        )
        mock_add_ip_route.assert_not_called()
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {}
        )

    @mock.patch.object(linux_net, "get_exposed_routes_on_network")
    @mock.patch.object(linux_net, "del_ip_route")
    def test__withdraw_subnet(
        self, mock_del_ip_route, mock_get_exposed_routes_on_network
    ):
        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        gateway["lrp_ports"] = {""}
        self.bgp_driver.ovn_local_cr_lrps = {"gateway_port": gateway}
        port_info = {
            "cr_lrp": self.cr_lrp0.logical_port,
            "subnets": {
                "fdcc:8cf2:d40c:2::/64",
                "192.168.1.0/24"
            }
        }

        mock_get_exposed_routes_on_network.side_effect = (["route-v4"], [])

        self.bgp_driver._withdraw_subnet(port_info, "gateway_port")

        expected_calls = [
            mock.call(
                mock.ANY,
                "192.168.1.0",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=24,
                via="10.0.0.10",
            ),
            mock.call(
                mock.ANY,
                "fdcc:8cf2:d40c:2::",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=64,
                via="fd51:f4b3:872:eda::10",
            ),
            mock.call(
                mock.ANY,
                "fd51:f4b3:872:eda::",
                CONF.bgp_vrf_table_id,
                CONF.bgp_nic,
                vlan=None,
                mask=64,
                via=None,
            ),
        ]

        mock_del_ip_route.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, "get_exposed_routes_on_network")
    @mock.patch.object(linux_net, "del_ip_route")
    def test__withdraw_subnet_no_gateway(
        self, mock_del_ip_route, mock_get_exposed_routes_on_network
    ):
        self.bgp_driver.ovn_local_cr_lrps = {}
        self.bgp_driver._withdraw_subnet(self.router_port, "gateway_port")
        mock_del_ip_route.assert_not_called()
        mock_get_exposed_routes_on_network.assert_not_called()

    @mock.patch.object(linux_net, "delete_ip_routes")
    @mock.patch.object(linux_net, "get_routes_on_tables")
    def test_sync(self, mock_get_routes_on_tables, mock_delete_ip_routes):
        def create_route(dst, dst_len, gateway):
            m = mock.Mock([])
            m.dst = dst
            m.dst_len = dst_len
            m.gateway = gateway
            return m

        def create_hashed_route(dst, dst_len, gateway):
            return ovn_stretched_l2_bgp_driver.HashedRoute(
                network=dst,
                prefix_len=dst_len,
                dst=gateway,
            )

        mock__expose_cr_lrp = mock.patch.object(
            self.bgp_driver, "_expose_cr_lrp"
        ).start()

        vrf_routes = [
            create_hashed_route(dst, dst_len, gateway)
            for (dst, dst_len, gateway) in [
                ("192.168.1.0", 24, "10.0.0.1"),
                ("10.0.0.0", 24, None),
                ("fdcc:8cf2:d40c:2::", 64, "fd51:f4b3:872:eda::1"),
                ("fd51:f4b3:872:eda::", 64, None),
            ]
        ]

        # really hacky way to get the routes into self.bgp_driver.vrf_routes
        mock__expose_cr_lrp.side_effect = (
            lambda _, __: self.bgp_driver.vrf_routes.update(vrf_routes)
        )

        self.sb_idl.get_cr_lrp_ports.return_value = [self.cr_lrp0]

        delete_route = create_route("192.168.0.0", 24, "10.0.0.1")
        routes = [
            create_route(dst, dst_len, gateway)
            for (dst, dst_len, gateway) in [
                ("192.168.1.0", 24, "10.0.0.1"),
                ("10.0.0.0", 24, None),
                ("fdcc:8cf2:d40c:2::", 64, "fd51:f4b3:872:eda::1"),
                ("fd51:f4b3:872:eda::", 64, None),
            ]
        ]
        routes.append(delete_route)

        mock_get_routes_on_tables.return_value = routes

        self.bgp_driver.sync()

        mock_get_routes_on_tables.assert_called_once_with(
            [CONF.bgp_vrf_table_id]
        )
        mock_delete_ip_routes.assert_called_once_with([delete_route])
        mock__expose_cr_lrp.assert_called_once_with(
            ["10.0.0.1/24", "fd51:f4b3:872:eda::1/64"], self.cr_lrp0
        )

    def test_withdraw_ip(self):
        mock__withdraw_cr_lrp = mock.patch.object(
            self.bgp_driver, "_withdraw_cr_lrp"
        ).start()

        self.bgp_driver.withdraw_ip(None, self.cr_lrp0)

        mock__withdraw_cr_lrp.assert_called_once_with(None, self.cr_lrp0)

    def test__withdraw_cr_lrp(self):
        mock__withdraw_subnet = mock.patch.object(
            self.bgp_driver, "_withdraw_subnet"
        ).start()

        gateway = {}
        gateway["ips"] = [
            ipaddress.ip_interface(ip)
            for ip in ["10.0.0.10/26", "fd51:f4b3:872:eda::10/64"]
        ]
        gateway["address_scopes"] = self.addr_scope
        self.bgp_driver.ovn_local_cr_lrps = {
            self.cr_lrp0.logical_port: gateway
        }
        gateway["lrp_ports"] = {
            "lrp-lrp_port0",
            "lrp-lrp_port1",
            "lrp-lrp_port2",
        }

        lrp_port0 = {
            "cr_lrp": self.cr_lrp0.logical_port,
            "subnets": {
                "fdcc:8cf2:d40c:1::/64",
                "192.168.0.0/24"}
        }
        lrp_port1 = {
            "cr_lrp": self.cr_lrp0.logical_port,
            "subnets": {
                "fdcc:8cf2:d40c:2::/64",
                "192.168.1.0/24"}
        }

        self.bgp_driver.propagated_lrp_ports = {
            "lrp-lrp_port0": lrp_port0,
            "lrp-lrp_port1": lrp_port1,
        }
        self.bgp_driver.ovn_local_cr_lrps = {
            self.cr_lrp0.logical_port: gateway
        }

        self.bgp_driver._withdraw_cr_lrp(None, self.cr_lrp0)
        mock__withdraw_subnet.assert_has_calls(
            [
                mock.call(lrp_port0, self.cr_lrp0.logical_port),
                mock.call(lrp_port1, self.cr_lrp0.logical_port),
            ],
            any_order=True
        )
        self.assertDictEqual(
            self.bgp_driver.ovn_local_cr_lrps,
            {}
        )
        self.assertDictEqual(
            self.bgp_driver.propagated_lrp_ports,
            {}
        )

    def test__withdraw_cr_lrp_invalid_addr_scope(self):
        mock__withdraw_subnet = mock.patch.object(
            self.bgp_driver, "_withdraw_subnet"
        ).start()

        gateway = {
            "address_scopes": {
                constants.IP_VERSION_4: '',
                constants.IP_VERSION_6: '',
            }
        }
        gateway["lrp_ports"] = {
            "lrp-lrp_port0",
            "lrp-lrp_port1",
            "lrp-lrp_port2",
        }
        self.bgp_driver.propagated_lrp_ports = {
            "lrp-lrp_port0": {},
            "lrp-lrp_port1": {},
            "lrp-lrp_port2": {},
        }
        self.bgp_driver.ovn_local_cr_lrps = {
            self.cr_lrp0.logical_port: gateway
        }

        self.bgp_driver._withdraw_cr_lrp(None, self.cr_lrp0)

        self.sb_idl.get_lrp_ports_for_router.assert_not_called()
        mock__withdraw_subnet.assert_not_called()

    def test_expose_ip(self):
        mock__expose_cr_lrp = mock.patch.object(
            self.bgp_driver, "_expose_cr_lrp"
        ).start()

        ips = ["10.0.0.1/24", "fd51:f4b3:872:eda::1/64"]

        self.bgp_driver.expose_ip(ips, self.cr_lrp0)

        mock__expose_cr_lrp.assert_called_once_with(ips, self.cr_lrp0)

    def test_expose_ip_invalid_type(self):
        mock__expose_cr_lrp = mock.patch.object(
            self.bgp_driver, "_expose_cr_lrp"
        ).start()

        patch_port = mock.Mock()
        patch_port.type = constants.OVN_PATCH_VIF_PORT_TYPE

        self.bgp_driver.expose_ip([], patch_port)

        mock__expose_cr_lrp.assert_not_called()

    def test__expose_cr_lrp(self):
        mock__ensure_network_exposed = mock.patch.object(
            self.bgp_driver, "_ensure_network_exposed"
        ).start()

        lrp_port0 = fakes.create_object(
            {
                "name": "fake-port-lrp0",
                "type": constants.OVN_PATCH_VIF_PORT_TYPE,
                "chassis": "fake-chassis1",
                "external_ids": {},
                "logical_port": "lrp-lrp_port0",
                "options": {
                    "chassis-redirect-port": self.cr_lrp0.logical_port,
                },
            }
        )

        lrp_port1 = fakes.create_object(
            {
                "name": "fake-port-lrp1",
                "type": constants.OVN_PATCH_VIF_PORT_TYPE,
                "mac": ["aa:bb:cc:dd:ee:ee 192.168.1.12 192.168.1.13"],
                "logical_port": "lrp-lrp_port1",
                "chassis": "fake-chassis1",
                "external_ids": {},
                "options": {},
            }
        )

        lrp_port2 = fakes.create_object(
            {
                "name": "fake-port-lrp2",
                "type": constants.OVN_PATCH_VIF_PORT_TYPE,
                "mac": [],
                "logical_port": "lrp-lrp_port2",
                "chassis": "fake-chassis1",
                "external_ids": {},
                "options": {},
            }
        )

        lrp_port3 = fakes.create_object(
            {
                "name": "fake-port-lrp3",
                "type": constants.OVN_PATCH_VIF_PORT_TYPE,
                "mac": [],
                "logical_port": "lrp-lrp_port3",
                "chassis": "",
                "up": [False],
                "external_ids": {},
                "options": {},
            }
        )

        lrp_port4 = fakes.create_object(
            {
                "name": "fake-port-lrp4",
                "type": constants.OVN_PATCH_VIF_PORT_TYPE,
                "mac": [],
                "logical_port": "lrp-lrp_port4",
                "chassis": "",
                "up": [False],
                "options": {},
            }
        )

        lrp_port5 = fakes.create_object(
            {
                "name": "fake-port-lrp5",
                "type": constants.OVN_PATCH_VIF_PORT_TYPE,
                "mac": [],
                "logical_port": "lrp-lrp_port5",
                "chassis": "",
                "up": [False],
                "options": {
                    "chassis-redirect-port": self.cr_lrp0.logical_port,
                },
            }
        )

        self.sb_idl.get_lrp_ports_for_router.return_value = [
            lrp_port0,
            lrp_port1,
            lrp_port2,
            lrp_port3,
            lrp_port4,
            lrp_port5,
        ]

        self.sb_idl.get_port_by_name.return_value = self.fake_patch_port

        ips = ["10.0.0.1/24", "fd51:f4b3:872:eda::1/64"]

        self.bgp_driver._expose_cr_lrp(ips, self.cr_lrp0)

        mock__ensure_network_exposed.assert_has_calls(
            [
                mock.call(lrp_port3, self.cr_lrp0.logical_port),
                mock.call(lrp_port4, self.cr_lrp0.logical_port),
            ]
        )

        self.sb_idl.get_port_by_name.assert_called_once_with("fake-port")

        self.assertEqual(
            {
                self.cr_lrp0.logical_port: {
                    "ips": [ipaddress.ip_interface(ip) for ip in ips],
                    "address_scopes": self.addr_scope,
                    "lrp_ports": set()
                }
            },
            self.bgp_driver.ovn_local_cr_lrps)

    def test__expose_cr_lrp_no_port(self):
        mock__ensure_network_exposed = mock.patch.object(
            self.bgp_driver, "_ensure_network_exposed"
        ).start()

        self.sb_idl.get_port_by_name.return_value = []

        ips = ["10.0.0.1/24", "fd51:f4b3:872:eda::1/64"]

        self.bgp_driver._expose_cr_lrp(ips, self.cr_lrp0)

        mock__ensure_network_exposed.assert_not_called()
        self.sb_idl.get_port_by_name.assert_called_once_with("fake-port")

    @mock.patch.object(driver_utils, "get_addr_scopes")
    def test__expose_cr_lrp_no_addr_scope(self, mock_addr_scopes):
        mock__ensure_network_exposed = mock.patch.object(
            self.bgp_driver, "_ensure_network_exposed"
        ).start()

        self.sb_idl.get_port_by_name.return_value = self.fake_patch_port

        mock_addr_scopes.return_value = {
            constants.IP_VERSION_4: "address_scope_v4",
            constants.IP_VERSION_6: "address_scope_v6",
        }

        self.bgp_driver._expose_cr_lrp([], self.cr_lrp0)

        self.sb_idl.get_port_by_name.assert_called_once_with("fake-port")
        mock_addr_scopes.assert_called_once_with(self.fake_patch_port)
        self.sb_idl.get_lrp_ports_for_router.assert_not_called()
        mock__ensure_network_exposed.assert_not_called()

    def test_expose_remote_ip(self):
        self.assertRaises(
            NotImplementedError,
            self.bgp_driver.expose_remote_ip,
            "1.2.3.4/24")

    def test_withdraw_remote_ip(self):
        self.assertRaises(
            NotImplementedError,
            self.bgp_driver.withdraw_remote_ip,
            "1.2.3.4/24")



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import nb_exceptions
from ovn_bgp_agent.drivers.openstack.utils import nat as nat_utils
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils as test_utils


class TestGetGatewayLrp(test_base.TestCase):
    def test_get(self):
        port = 'foo'
        nat = test_utils.create_row(gateway_port=[port])
        observed = nat_utils.get_gateway_lrp(nat)
        self.assertEqual(port, observed)

    def test_no_gw_port(self):
        nat = test_utils.create_row(gateway_port=[])
        self.assertRaises(
            nb_exceptions.NATNotFound, nat_utils.get_gateway_lrp, nat)


class TestGetChassisHostingCrlrp(test_base.TestCase):
    def test_get_chassis(self):
        chassis = 'foo'
        port = test_utils.create_row(
            status={constants.OVN_STATUS_CHASSIS: chassis})
        nat = test_utils.create_row(gateway_port=[port])
        observed = nat_utils.get_chassis_hosting_crlrp(nat)
        self.assertEqual(chassis, observed)

    def test_no_chasssis(self):
        port = test_utils.create_row(
            status={})
        nat = test_utils.create_row(gateway_port=[port])
        self.assertRaises(
            nb_exceptions.ChassisNotFound,
            nat_utils.get_chassis_hosting_crlrp, nat)



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import common as common_utils


def get_name_from_external_ids(row, key=constants.OVN_LB_LR_REF_EXT_ID_KEY):
    router_name = common_utils.get_from_external_ids(row, key)

    try:
        return router_name.replace('neutron-', "", 1)
    except AttributeError:
        pass



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_privsep import capabilities
from oslo_privsep import priv_context

default = priv_context.PrivContext(
    __name__,
    cfg_section='privsep',
    pypath=__name__ + '.default',
    capabilities=[capabilities.CAP_DAC_OVERRIDE,
                  capabilities.CAP_DAC_READ_SEARCH,
                  capabilities.CAP_NET_ADMIN,
                  capabilities.CAP_SYS_ADMIN],
)

ovs_vsctl_cmd = priv_context.PrivContext(
    __name__,
    cfg_section='privsep_ovs_vsctl',
    pypath=__name__ + '.ovs_vsctl_cmd',
    capabilities=[capabilities.CAP_SYS_ADMIN,
                  capabilities.CAP_NET_ADMIN]
)

vtysh_cmd = priv_context.PrivContext(
    __name__,
    cfg_section='privsep_vtysh',
    pypath=__name__ + '.vtysh_cmd',
    capabilities=[capabilities.CAP_SYS_ADMIN,
                  capabilities.CAP_NET_ADMIN]
)



# Copyright 2022 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ipaddress
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.utils import linux_net

LOG = logging.getLogger(__name__)


def is_ipv6_gua(ip):
    if linux_net.get_ip_version(ip) != constants.IP_VERSION_6:
        return False

    ipv6 = ipaddress.IPv6Address(ip.split('/')[0])
    if ipv6.is_global:
        return True
    return False


def get_addr_scopes(port):
    return {
        constants.IP_VERSION_4: port.external_ids.get(
            constants.SUBNET_POOL_ADDR_SCOPE4),
        constants.IP_VERSION_6: port.external_ids.get(
            constants.SUBNET_POOL_ADDR_SCOPE6),
        }


def get_port_chassis(port, chassis,
                     default_port_type=constants.OVN_VM_VIF_PORT_TYPE):
    # row.options['requested-chassis'] superseeds the id in external_ids.
    # Since it is not used for virtual ports by ovn, this option will be
    # ignored for virtual ports.

    # since 'old' rows could be used, it will not hold the type information
    # if that is the case, please supply a default in the arguments.

    port_type = getattr(port, 'type', default_port_type)
    if (port_type != constants.OVN_VIRTUAL_VIF_PORT_TYPE and
            hasattr(port, 'options') and
            port.options.get(constants.OVN_REQUESTED_CHASSIS)):

        # requested-chassis can be a comma separated list,
        # so lets only return our chassis if it is a list, to be able
        # to do a == equal comparison
        req_chassis = port.options[constants.OVN_REQUESTED_CHASSIS]
        if chassis in req_chassis.split(','):
            return chassis

        return req_chassis.split(',')[0]

    elif (hasattr(port, 'external_ids') and
            port.external_ids.get(constants.OVN_HOST_ID_EXT_ID_KEY)):
        return port.external_ids[constants.OVN_HOST_ID_EXT_ID_KEY]


def check_name_prefix(entity, prefix):
    try:
        return entity.name.startswith(prefix)
    except AttributeError:
        return False


def is_pf_lb(lb):
    return check_name_prefix(lb, constants.OVN_LB_PF_NAME_PREFIX)


def ips_per_version(ips: 'list[str]') -> 'dict[int, str]':
    '''Separate list of ips into ip versions.

    For example, this list ['10.0.0.1/32', 'fe80::1/128'] will be converted
    to dictionary {
        4: '10.0.0.1',
        6: 'fe80::1',
    }

    If there are more than 1 ip for the same ip version, it will overwrite
    the previous ip for that ip version.
    '''
    ip_list = {constants.IP_VERSION_4: None,
               constants.IP_VERSION_6: None}

    for ip in ips:
        ver = linux_net.get_ip_version(ip)
        ip_list[ver] = ipaddress.ip_address(ip.split('/')[0]).compressed

    return ip_list


def get_prefixes_from_ips(ips: 'list[str]') -> 'list[str]':
    '''Return the network address for any given ip (with mask)

    For a list like ['192.168.0.1/24'] it will return ['192.168.0.0/24']
    '''
    return ['/'.join([ipaddress.ip_network(ip, strict=False)[0].compressed,
                      ip.split('/')[-1]])
            for ip in ips]


def remove_port_from_ip(ip_address):
    last_colon_index = ip_address.rfind(':')
    # no port
    if last_colon_index == -1:
        return ip_address
    # check if right side from index is a digit, in positive case remove it.
    # For IPv6 it will come on format [ipv6]:port, so will also remove
    # correctly just only the port
    if ip_address[last_colon_index + 1:].isdigit():
        return ip_address[:last_colon_index]
    return ip_address


def get_port_vlan(port):
    '''Will return the tag of the given logical switch port row as string

    If the vlan tag is not configured, it will return the value of
    constants.VLAN_ID_UNTAGGED
    '''
    if port.tag:
        return str(port.tag[0])
    return str(constants.VLAN_ID_UNTAGGED)


def get_port_vrf_settings(port):
    """Create a comparable object with the settings of the vrf

    Returns None if the settings are not found in the port, otherwise it
    will return a string value. Either an empty string or a concatenation
    of the type and vni, e.g. l3::1001
    """
    if hasattr(port, 'external_ids'):
        try:
            return "%s::%s" % (
                port.external_ids[constants.OVN_EVPN_TYPE_EXT_ID_KEY],
                port.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY])
        except (AttributeError, KeyError):
            return ''



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import ipaddress

from unittest import mock

from ovn_bgp_agent import constants
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.utils import linux_net


class IPRouteDict(dict):
    def get_attr(self, attr_name):
        for attr in self['attrs']:
            if attr[0] == attr_name:
                return attr[1]
        return


class TestLinuxNet(test_base.TestCase):

    def setUp(self):
        super(TestLinuxNet, self).setUp()
        # Mock pyroute2.NDB context manager object
        self.mock_ndb = mock.patch.object(linux_net.pyroute2, 'NDB').start()
        self.fake_ndb = self.mock_ndb().__enter__()

        # Mock pyroute2.IPRoute context manager object
        self.mock_ipr = mock.patch.object(linux_net.pyroute2,
                                          'IPRoute').start()
        self.fake_ipr = self.mock_ipr().__enter__()

        # Helper variables used accross many tests
        self.ip = '10.10.1.16'
        self.ipv6 = '2002::1234:abcd:ffff:c0a8:101'
        self.dev = 'ethfake'
        self.mac = 'aa:bb:cc:dd:ee:ff'
        self.bridge = 'br-fake'
        self.table_id = 100
        self.network = ipaddress.IPv4Network("10.10.1.0/24")
        self.network_v6 = ipaddress.IPv6Network("2002:0:0:1234:0:0:0:0/64")

    def test_get_ip_version_v4(self):
        self.assertEqual(4, linux_net.get_ip_version('%s/32' % self.ip))
        self.assertEqual(4, linux_net.get_ip_version(self.ip))

    def test_get_ip_version_v6(self):
        self.assertEqual(6, linux_net.get_ip_version('%s/64' % self.ipv6))
        self.assertEqual(6, linux_net.get_ip_version(self.ipv6))

    def test_get_interfaces(self):
        iface0 = IPRouteDict({'attrs': [('IFLA_IFNAME', 'ethfake0')]})
        iface1 = IPRouteDict({'attrs': [('IFLA_IFNAME', 'ethfake1')]})
        iface2 = IPRouteDict({'attrs': [('IFLA_IFNAME', 'ethfake2')]})
        self.fake_ipr.get_links.return_value = [iface0, iface1, iface2]

        ret = linux_net.get_interfaces(filter_out='ethfake1')
        self.assertEqual(['ethfake0', 'ethfake2'], ret)

    def test_get_interface_index(self):
        self.fake_ipr.link_lookup.return_value = [7]
        ret = linux_net.get_interface_index('fake-nic')
        self.assertEqual(7, ret)

    def test_get_interface_index_error(self):
        self.fake_ipr.link_lookup.return_value = ''
        self.assertRaises(agent_exc.NetworkInterfaceNotFound,
                          linux_net.get_interface_index, 'fake-nic')

    def test_get_interface_address(self):
        device_idx = 7
        self.fake_ipr.link_lookup.return_value = [device_idx]
        fake_link = mock.MagicMock()
        fake_link.get_attr.return_value = self.mac
        self.fake_ipr.get_links.return_value = [fake_link]

        ret = linux_net.get_interface_address('fake-nic')
        self.assertEqual(self.mac, ret)

    def test_get_interface_address_index_error(self):
        self.fake_ipr.link_lookup.return_value = ''
        self.assertRaises(agent_exc.NetworkInterfaceNotFound,
                          linux_net.get_interface_address, 'fake-nic')

    def test_get_nic_info(self):
        device_idx = 7
        nic_addr = IPRouteDict({'prefixlen': 32,
                                'attrs': [('IFA_ADDRESS', self.ip)]})
        self.fake_ipr.link_lookup.return_value = [device_idx]
        self.fake_ipr.get_addr.return_value = [nic_addr]
        fake_link = mock.MagicMock()
        fake_link.get_attr.return_value = self.mac
        self.fake_ipr.get_links.return_value = [fake_link]

        ret = linux_net.get_nic_info('fake-nic')
        self.assertEqual(('{}/32'.format(self.ip), self.mac), ret)

    def test_get_nic_info_index_error(self):
        self.fake_ipr.link_lookup.return_value = ''
        self.assertRaises(agent_exc.NetworkInterfaceNotFound,
                          linux_net.get_nic_info, 'fake-nic')

    @mock.patch('ovn_bgp_agent.privileged.linux_net.ensure_vrf')
    def test_ensure_vrf(self, mock_ensure_vrf):
        linux_net.ensure_vrf('fake-vrf', 10)
        mock_ensure_vrf.assert_called_once_with('fake-vrf', 10)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.ensure_bridge')
    def test_ensure_bridge(self, mock_ensure_bridge):
        linux_net.ensure_bridge('fake-bridge')
        mock_ensure_bridge.assert_called_once_with('fake-bridge')

    @mock.patch('ovn_bgp_agent.privileged.linux_net.ensure_vxlan')
    def test_ensure_vxlan(self, mock_ensure_vxlan):
        linux_net.ensure_vxlan('fake-vxlan', 11, self.ip, 7)
        mock_ensure_vxlan.assert_called_once_with('fake-vxlan', 11, self.ip, 7)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.ensure_veth')
    def test_ensure_veth(self, mock_ensure_veth):
        linux_net.ensure_veth('fake-veth', 'fake-veth-peer')
        mock_ensure_veth.assert_called_once_with('fake-veth', 'fake-veth-peer')

    @mock.patch('ovn_bgp_agent.privileged.linux_net.ensure_dummy_device')
    def test_ensure_dummy_device(self, mock_ensure_dummy_device):
        linux_net.ensure_dummy_device('fake-dev')
        mock_ensure_dummy_device.assert_called_once_with('fake-dev')

    @mock.patch.object(linux_net, 'ensure_dummy_device')
    @mock.patch.object(linux_net, 'set_master_for_device')
    def test_ensure_ovn_device(self, mock_master, mock_dummy):
        linux_net.ensure_ovn_device('ifname', 'fake-vrf')
        mock_dummy.assert_called_once_with('ifname')
        mock_master.assert_called_once_with('ifname', 'fake-vrf')

    @mock.patch('ovn_bgp_agent.privileged.linux_net.delete_device')
    def test_delete_device(self, mock_delete_device):
        linux_net.delete_device('fake-dev')
        mock_delete_device.assert_called_once_with('fake-dev')

    @mock.patch.object(linux_net, 'enable_proxy_arp')
    @mock.patch.object(linux_net, 'enable_proxy_ndp')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.add_ip_to_dev')
    def test_ensure_arp_ndp_enabled_for_bridge(self, mock_add_ip_to_dev,
                                               mock_ndp, mock_arp):
        linux_net.ensure_arp_ndp_enabled_for_bridge('fake-bridge', 511)
        # NOTE(ltomasbo): hardoced starting ipv4 is 192.168.0.0, and ipv6 is
        # fd53:d91e:400:7f17::0
        ipv4 = '169.254.1.255'  # base + 511 offset
        ipv6 = 'fd53:d91e:400:7f17::1ff'  # base + 5122 offset (to hex)
        calls = [mock.call(ipv4, 'fake-bridge'),
                 mock.call(ipv6, 'fake-bridge')]
        mock_add_ip_to_dev.assert_has_calls(calls)
        mock_ndp.assert_called_once_with('fake-bridge')
        mock_arp.assert_called_once_with('fake-bridge')

    @mock.patch.object(linux_net, 'enable_proxy_arp')
    @mock.patch.object(linux_net, 'enable_proxy_ndp')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.add_ip_to_dev')
    def test_ensure_arp_ndp_enabled_for_bridge_vlan(self, mock_add_ip_to_dev,
                                                    mock_ndp, mock_arp):
        linux_net.ensure_arp_ndp_enabled_for_bridge('fake-bridge', 511, 11)
        # NOTE(ltomasbo): hardoced starting ipv4 is 192.168.0.0, and ipv6 is
        # fd53:d91e:400:7f17::0
        ipv4 = '169.254.1.255'  # base + 511 offset
        ipv6 = 'fd53:d91e:400:7f17::1ff'  # base + 5122 offset (to hex)
        calls = [mock.call(ipv4, 'fake-bridge'),
                 mock.call(ipv6, 'fake-bridge')]
        mock_add_ip_to_dev.assert_has_calls(calls)
        mock_ndp.assert_called_once_with('fake-bridge')
        mock_arp.assert_called_once_with('fake-bridge')

    @mock.patch.object(linux_net, 'enable_proxy_arp')
    @mock.patch.object(linux_net, 'enable_proxy_ndp')
    @mock.patch(
        'ovn_bgp_agent.privileged.linux_net.ensure_vlan_device_for_network')
    def test_ensure_vlan_device_for_network(
            self, mock_ensure_vlan_device_for_network, mock_ndp, mock_arp):
        linux_net.ensure_vlan_device_for_network('fake-br', 10)
        expected_dev = 'fake-br/10'
        mock_ensure_vlan_device_for_network.assert_called_once_with(
            'fake-br', 10)
        mock_ndp.assert_called_once_with(expected_dev)
        mock_arp.assert_called_once_with(expected_dev)

    @mock.patch.object(linux_net, 'enable_proxy_arp')
    @mock.patch.object(linux_net, 'enable_proxy_ndp')
    @mock.patch(
        'ovn_bgp_agent.privileged.linux_net.ensure_vlan_device_for_network')
    def test_ensure_vlan_trimmed_device_for_network(
            self, mock_ensure_vlan_device_for_network, mock_ndp, mock_arp):
        linux_net.ensure_vlan_device_for_network('fake-provider', 1020)
        mock_ensure_vlan_device_for_network.assert_called_once_with(
            'fake-provider', 1020)
        expected_dev = 'fake-provi/1020'
        mock_ndp.assert_called_once_with(expected_dev)
        mock_arp.assert_called_once_with(expected_dev)

    @mock.patch.object(linux_net, 'delete_device')
    def test_delete_vlan_device_for_network(self, mock_del):
        linux_net.delete_vlan_device_for_network('fake-br', 10)
        vlan_name = 'fake-br.10'
        mock_del.assert_called_once_with(vlan_name)

    @mock.patch.object(linux_net, 'delete_device')
    def test_delete_vlan_trimmed_device_for_network(self, mock_del):
        linux_net.delete_vlan_device_for_network('fake-provider', 1020)
        vlan_name = 'fake-provi.1020'
        mock_del.assert_called_once_with(vlan_name)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.set_kernel_flag')
    def test_enable_proxy_ndp(self, mock_flag):
        linux_net.enable_proxy_ndp(self.dev)
        expected_flag = 'net.ipv6.conf.%s.proxy_ndp' % self.dev
        mock_flag.assert_called_once_with(expected_flag, 1)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.set_kernel_flag')
    def test_enable_proxy_arp(self, mock_flag):
        linux_net.enable_proxy_arp(self.dev)
        expected_flag = 'net.ipv4.conf.%s.proxy_arp' % self.dev
        mock_flag.assert_called_once_with(expected_flag, 1)

    def test_get_exposed_ips(self):
        ip0 = IPRouteDict({'prefixlen': 32,
                           'attrs': [('IFA_ADDRESS', self.ip)]})
        ip1 = IPRouteDict({'prefixlen': 128,
                           'attrs': [('IFA_ADDRESS', self.ipv6)]})
        ip2 = IPRouteDict({'prefixlen': 24,
                           'attrs': [('IFA_ADDRESS', '10.10.1.18')]})
        ip3 = IPRouteDict(
            {'prefixlen': 64,
             'attrs': [('IFA_ADDRESS', '2001:0DB8:0000:000b::')]})
        self.fake_ipr.get_addr.return_value = [ip0, ip1, ip2, ip3]

        ips = linux_net.get_exposed_ips(self.dev)

        expected_ips = [self.ip, self.ipv6]
        self.assertEqual(expected_ips, ips)

    def test_get_nic_ip(self):
        ip0 = IPRouteDict({'attrs': [('IFA_ADDRESS', '10.10.1.16')]})
        ip1 = IPRouteDict({'attrs': [('IFA_ADDRESS', '10.10.1.17')]})
        self.fake_ipr.get_addr.return_value = [ip0, ip1]

        ips = linux_net.get_nic_ip(self.dev)

        expected_ips = ['10.10.1.16', '10.10.1.17']
        self.assertEqual(expected_ips, ips)

    def test_get_exposed_ips_on_network(self):
        ip0 = IPRouteDict({'prefixlen': 32,
                           'attrs': [('IFA_ADDRESS', self.ip)]})
        ip1 = IPRouteDict({'prefixlen': 128,
                           'attrs': [('IFA_ADDRESS', '10.10.1.17')]})
        ip2 = IPRouteDict({'prefixlen': 128,
                           'attrs': [('IFA_ADDRESS', self.ipv6)]})
        ip3 = IPRouteDict({
            'prefixlen': 128,
            'attrs': [('IFA_ADDRESS', '2001:db8:3333:4444:5555:6666:7777:8888')
                      ]})

        self.fake_ipr.get_addr.return_value = [ip0, ip1, ip2, ip3]

        network_ips = [ipaddress.ip_address(self.ip),
                       ipaddress.ip_address(self.ipv6)]
        ret = linux_net.get_exposed_ips_on_network(self.dev, network_ips)

        self.assertEqual([self.ip, self.ipv6], ret)

    def test_get_exposed_routes_on_network_v4(self):
        route0 = mock.MagicMock(
            dst=mock.Mock(),
            table=self.table_id,
            scope=1,
            proto=11,
            gateway=self.ip,
        )
        route1 = mock.MagicMock(
            dst=mock.Mock(),
            table=self.table_id,
            scope=1,
            proto=11,
            gateway=self.ipv6,
        )
        route2 = mock.MagicMock(
            dst=mock.Mock(),
            table=self.table_id,
            scope=1,
            proto=11,
            gateway=None,
        )

        self.fake_ndb.routes.dump.return_value = [route0, route1, route2]
        ret = linux_net.get_exposed_routes_on_network(
            [self.table_id], self.network
        )

        self.assertEqual([route0], ret)

    def test_get_exposed_routes_on_network_v6(self):
        route0 = mock.MagicMock(
            dst=mock.Mock(),
            table=self.table_id,
            scope=1,
            proto=11,
            gateway=self.ip,
        )
        route1 = mock.MagicMock(
            dst=mock.Mock(),
            table=self.table_id,
            scope=1,
            proto=11,
            gateway=self.ipv6,
        )
        route2 = mock.MagicMock(
            dst=mock.Mock(),
            table=self.table_id,
            scope=1,
            proto=11,
            gateway=None,
        )

        self.fake_ndb.routes.dump.return_value = [route0, route1, route2]
        ret = linux_net.get_exposed_routes_on_network(
            [self.table_id], self.network_v6
        )

        self.assertEqual([route1], ret)

    def test_get_ovn_ip_rules(self):
        rule0 = IPRouteDict({'dst_len': 128, 'family': 10,
                             'attrs': [('FRA_TABLE', 7),
                                       ('FRA_DST', 10)]})
        rule1 = IPRouteDict({'dst_len': 32, 'family': 2,
                             'attrs': [('FRA_TABLE', 7),
                                       ('FRA_DST', 11)]})
        rule2 = IPRouteDict({'dst_len': 24, 'family': 2,
                             'attrs': [('FRA_TABLE', 9),
                                       ('FRA_DST', 5)]})
        rule3 = IPRouteDict({'dst_len': 128, 'family': 10,
                             'attrs': [('FRA_TABLE', 10),
                                       ('FRA_DST', 6)]})
        self.fake_ipr.get_rules.side_effect = [[rule1, rule2], [rule0, rule3]]

        ret = linux_net.get_ovn_ip_rules([7, 10])
        expected_ret = {'10/128': {'table': 7, 'family': 10},
                        '11/32': {'table': 7, 'family': 2},
                        '6/128': {'table': 10, 'family': 10}}
        self.assertEqual(expected_ret, ret)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.delete_exposed_ips')
    def test_delete_exposed_ips(self, mock_delete_exposed_ips):
        linux_net.delete_exposed_ips([self.ip], self.dev)
        mock_delete_exposed_ips.assert_called_once_with([self.ip], self.dev)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.delete_ip_rules')
    def test_delete_ip_rules(self, mock_delete_ip_rules):
        ip_rules = {'10/128': {'table': 7, 'family': 'fake'},
                    '6/128': {'table': 10, 'family': 'fake'}}
        linux_net.delete_ip_rules(ip_rules)
        mock_delete_ip_rules.assert_called_once_with(ip_rules)

    @mock.patch.object(linux_net, 'get_interface_index')
    def _test_delete_bridge_ip_routes(self, mock_route_delete, mock_get_index,
                                      is_vlan=False, has_gateway=False,
                                      trimmed_bridge=False):
        gateway = '1.1.1.1'
        oif = 11
        bridge = 'br-provider' if trimmed_bridge else self.bridge
        vlan = 2030 if is_vlan else None
        mock_get_index.return_value = oif

        route = {'route': {'dst': self.ip,
                           'dst_len': 32,
                           'table': 20},
                 'vlan': vlan}
        if has_gateway:
            route['route']['gateway'] = gateway

        routing_tables = {bridge: 20}
        routing_tables_routes = {bridge: [route]}
        # extra_route0 matches with the route
        extra_route0 = IPRouteDict({
            'dst_len': 32, 'family': constants.AF_INET, 'table': 20,
            'attrs': [('RTA_DST', self.ip),
                      ('RTA_OIF', oif),
                      ('RTA_GATEWAY', gateway)]})
        # extra_route1 does not match with route and should be removed
        extra_route1 = IPRouteDict({
            'dst_len': 32, 'family': constants.AF_INET, 'table': 20,
            'attrs': [('RTA_DST', '10.10.1.17'),
                      ('RTA_OIF', oif),
                      ('RTA_GATEWAY', gateway)]})
        extra_routes = {bridge: [extra_route0, extra_route1]}

        linux_net.delete_bridge_ip_routes(
            routing_tables, routing_tables_routes, extra_routes)

        # Assert extra_route1 has been removed
        expected_route = {'dst': '10.10.1.17', 'dst_len': 32,
                          'family': constants.AF_INET, 'oif': oif,
                          'gateway': gateway, 'table': 20}

        mock_route_delete.assert_called_once_with(expected_route)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_delete_bridge_ip_routes(self, mock_route_delete):
        self._test_delete_bridge_ip_routes(mock_route_delete)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_delete_bridge_ip_routes_vlan(self, mock_route_delete):
        self._test_delete_bridge_ip_routes(mock_route_delete, is_vlan=True)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_delete_trimmed_bridge_ip_routes_vlan(self, mock_route_delete):
        self._test_delete_bridge_ip_routes(mock_route_delete,
                                           is_vlan=True,
                                           trimmed_bridge=True)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_delete_bridge_ip_routes_gateway(self, mock_route_delete):
        self._test_delete_bridge_ip_routes(mock_route_delete, has_gateway=True)

    @mock.patch('ovn_bgp_agent.utils.linux_net.delete_ip_routes')
    def test_delete_routes_from_table(self, mock_delete_ip_routes):
        route0 = {'scope': 1, 'proto': 11}
        route1 = {'scope': 2, 'proto': 22}
        route2 = {'scope': 254, 'proto': 186}
        self.fake_ipr.get_routes.return_value = [
            route0, route1, route2]

        linux_net.delete_routes_from_table('fake-table')

        mock_delete_ip_routes.assert_called_once_with([route0, route1])

    def test_get_routes_on_tables(self):
        route0 = IPRouteDict({
            'proto': 10, 'table': 10,
            'attrs': [('RTA_DST', '10.10.10.10')]})
        # Route1 has proto 186, should be ignored
        route1 = IPRouteDict({
            'proto': 186, 'table': 11,
            'attrs': [('RTA_DST', '11.11.11.11')]})
        route2 = IPRouteDict({
            'proto': 12, 'table': 11,
            'attrs': [('RTA_DST', '12.12.12.12')]})
        # Route3 is in the list but dst is empty
        route3 = IPRouteDict({
            'proto': 10, 'table': 22,
            'attrs': [('RTA_DST', '')]})
        self.fake_ipr.get_routes.side_effect = [
            [route0], [route1, route2], [route3]]

        ret = linux_net.get_routes_on_tables([10, 11, 22])

        self.assertEqual([route0, route2], ret)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_delete_ip_routes(self, mock_route_delete):
        route0 = dict(
            table=10, dst='10.10.10.10', proto=10, dst_len=128,
            oif='ethout', family='fake', gateway='1.1.1.1')
        route1 = dict(
            table=11, dst='11.11.11.11', proto=11, dst_len=64,
            oif='ethout', family='fake', gateway='2.2.2.2')
        routes = [route0, route1]

        linux_net.delete_ip_routes(routes)

        route0.pop('proto')
        route1.pop('proto')
        mock_route_delete.assert_has_calls(
            [mock.call(route0), mock.call(route1)])

    @mock.patch('ovn_bgp_agent.privileged.linux_net.add_ndp_proxy')
    def test_add_ndp_proxy(self, mock_ndp_proxy):
        linux_net.add_ndp_proxy(self.ip, self.dev, vlan=10)
        mock_ndp_proxy.assert_called_once_with(self.ip, self.dev, 10)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.del_ndp_proxy')
    def test_del_ndp_proxy(self, mock_ndp_proxy):
        linux_net.del_ndp_proxy(self.ip, self.dev, vlan=10)
        mock_ndp_proxy.assert_called_once_with(self.ip, self.dev, 10)

    @mock.patch.object(linux_net, 'get_interface_index')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.add_ip_to_dev')
    def test_add_ips_to_dev(self, mock_add_ip_to_dev, mock_route_delete,
                            mock_get_index):
        ips = [self.ip, self.ipv6]
        oif = 7
        mock_get_index.return_value = oif
        linux_net.add_ips_to_dev(
            self.dev, ips, clear_local_route_at_table=123)

        # Assert called for each ip
        calls = [mock.call(self.ip, self.dev),
                 mock.call(self.ipv6, self.dev)]
        mock_add_ip_to_dev.assert_has_calls(calls)

        r1 = {'table': 123, 'proto': 2, 'scope': 254, 'dst': self.ip,
              'oif': oif}
        r2 = {'table': 123, 'proto': 2, 'scope': 254, 'dst': self.ipv6,
              'oif': oif}
        calls = [mock.call(r1),
                 mock.call(r2)]
        mock_route_delete.assert_has_calls(calls)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.del_ip_from_dev')
    def test_del_ips_from_dev(self, mock_del_ip_from_dev):
        ips = [self.ip, self.ipv6]
        linux_net.del_ips_from_dev(self.dev, ips)

        calls = [mock.call(self.ip, self.dev),
                 mock.call(self.ipv6, self.dev)]
        mock_del_ip_from_dev.assert_has_calls(calls)

    @mock.patch.object(linux_net, 'add_ip_nei')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.rule_create')
    def test_add_ip_rule(self, mock_rule_create, mock_add_ip_nei):
        linux_net.add_ip_rule(
            self.ip, 7, dev=self.dev, lladdr=self.mac)

        expected_args = {'dst': self.ip, 'table': 7,
                         'dst_len': 32, 'family': constants.AF_INET}
        mock_rule_create.assert_called_once_with(expected_args)
        mock_add_ip_nei.assert_called_once_with(self.ip, self.mac, self.dev)

    @mock.patch.object(linux_net, 'add_ip_nei')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.rule_create')
    def test_add_ip_rule_ipv6(self, mock_rule_create, mock_add_ip_nei):
        linux_net.add_ip_rule(self.ipv6, 7, dev=self.dev, lladdr=self.mac)

        expected_args = {'dst': self.ipv6,
                         'table': 7, 'dst_len': 128,
                         'family': constants.AF_INET6}
        mock_rule_create.assert_called_once_with(expected_args)
        mock_add_ip_nei.assert_called_once_with(self.ipv6, self.mac, self.dev)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.rule_create')
    def test_add_ip_rule_invalid_ip(self, mock_rule_create):
        self.assertRaises(agent_exc.InvalidPortIP,
                          linux_net.add_ip_rule, '10.10.1.6/30/128', 7)
        mock_rule_create.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.add_ip_nei')
    def test_add_ip_nei(self, mock_add_ip_nei):
        linux_net.add_ip_nei(self.ip, self.mac, self.dev)
        mock_add_ip_nei.assert_called_once_with(self.ip, self.mac, self.dev)

    @mock.patch.object(linux_net, 'del_ip_nei')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.rule_delete')
    def test_del_ip_rule(self, mock_rule_delete, mock_del_ip_nei):
        linux_net.del_ip_rule(self.ip, 7, dev=self.dev, lladdr=self.mac)

        expected_args = {'dst': self.ip, 'table': 7,
                         'dst_len': 32, 'family': constants.AF_INET}
        mock_rule_delete.assert_called_once_with(expected_args)
        mock_del_ip_nei.assert_called_once_with(self.ip, self.mac, self.dev)

    @mock.patch.object(linux_net, 'del_ip_nei')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.rule_delete')
    def test_del_ip_rule_ipv6(self, mock_rule_delete, mock_del_ip_nei):
        linux_net.del_ip_rule(self.ipv6, 7, dev=self.dev, lladdr=self.mac)

        expected_args = {'dst': self.ipv6, 'table': 7,
                         'dst_len': 128, 'family': constants.AF_INET6}
        mock_rule_delete.assert_called_once_with(expected_args)
        mock_del_ip_nei.assert_called_once_with(self.ipv6, self.mac, self.dev)

    @mock.patch.object(linux_net, 'del_ip_nei')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.rule_delete')
    def test_del_ip_rule_invalid_ip(self, mock_rule_delete, mock_del_ip_nei):
        self.assertRaises(agent_exc.InvalidPortIP,
                          linux_net.del_ip_rule, '10.10.1.6/30/128', 7)

        mock_rule_delete.assert_not_called()
        mock_del_ip_nei.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.del_ip_nei')
    def test_del_ip_nei(self, mock_del_ip_nei):
        linux_net.del_ip_nei(self.ip, self.mac, self.dev)
        mock_del_ip_nei.assert_called_once_with(self.ip, self.mac, self.dev)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.add_unreachable_route')
    def test_add_unreachable_route(self, mock_add_route):
        linux_net.add_unreachable_route('fake-vrf')
        mock_add_route.assert_called_once_with('fake-vrf')

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route(self, mock_route_create):
        routes = {}
        linux_net.add_ip_route(routes, self.ip, 7, self.dev)
        expected_routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': None}]}
        self.assertEqual(expected_routes, routes)
        mock_route_create.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route_ipv6(self, mock_route_create):
        routes = {}
        linux_net.add_ip_route(routes, self.ipv6, 7, self.dev)
        expected_routes = {
            self.dev: [{'route': {'dst': self.ipv6,
                                  'dst_len': 128,
                                  'family': constants.AF_INET6,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'table': 7},
                        'vlan': None}]}
        self.assertEqual(expected_routes, routes)
        mock_route_create.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route_via(self, mock_route_create):
        routes = {}
        linux_net.add_ip_route(routes, self.ip, 7, self.dev, via='1.1.1.1')
        expected_routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'gateway': '1.1.1.1',
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 0,
                                  'table': 7},
                        'vlan': None}]}
        self.assertEqual(expected_routes, routes)
        mock_route_create.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route_vlan(self, mock_route_create):
        routes = {}
        linux_net.add_ip_route(routes, self.ip, 7, self.dev, vlan=10)
        expected_routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': 10}]}
        self.assertEqual(expected_routes, routes)
        mock_route_create.assert_not_called()

    @mock.patch.object(linux_net, 'get_interface_index')
    @mock.patch.object(linux_net, 'ensure_vlan_device_for_network')
    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route_vlan_keyerror(self, mock_route_create,
                                        mock_ensure_vlan_device,
                                        mock_get_index):
        routes = {}
        oif = '5'
        mock_get_index.side_effect = [agent_exc.NetworkInterfaceNotFound, oif]
        linux_net.add_ip_route(routes, self.ip, 7, self.dev, vlan=10)
        expected_routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': oif,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': 10}]}
        self.assertEqual(expected_routes, routes)
        mock_ensure_vlan_device.assert_called_once_with(self.dev, 10)
        mock_route_create.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route_mask(self, mock_route_create):
        routes = {}
        linux_net.add_ip_route(routes, self.ip, 7, self.dev, mask=30)
        expected_routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 30,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': None}]}
        self.assertEqual(expected_routes, routes)
        mock_route_create.assert_not_called()

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_create')
    def test_add_ip_route_no_route(self, mock_route_create):
        self.fake_ipr.route.return_value = ()
        routes = {}
        linux_net.add_ip_route(routes, self.ip, 7, self.dev)
        expected_routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': None}]}
        self.assertEqual(expected_routes, routes)
        mock_route_create.assert_called_once_with(
            expected_routes[self.dev][0]['route'])

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_del_ip_route(self, mock_route_delete):
        routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': None}]}
        route = copy.deepcopy(routes[self.dev][0]['route'])

        linux_net.del_ip_route(routes, self.ip, 7, self.dev)

        self.assertEqual({self.dev: []}, routes)
        mock_route_delete.assert_called_once_with(route)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_del_ip_route_ipv6(self, mock_route_delete):
        routes = {
            self.dev: [{'route': {'dst': self.ipv6,
                                  'dst_len': 128,
                                  'family': constants.AF_INET6,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'table': 7},
                        'vlan': None}]}
        route = copy.deepcopy(routes[self.dev][0]['route'])

        linux_net.del_ip_route(routes, self.ipv6, 7, self.dev)

        self.assertEqual({self.dev: []}, routes)
        mock_route_delete.assert_called_once_with(route)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_del_ip_route_via(self, mock_route_delete):
        routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': mock.ANY,
                                  'gateway': '1.1.1.1',
                                  'proto': 3,
                                  'scope': 0,
                                  'table': 7},
                        'vlan': None}]}
        route = copy.deepcopy(routes[self.dev][0]['route'])

        linux_net.del_ip_route(routes, self.ip, 7, self.dev, via='1.1.1.1')

        self.assertEqual({self.dev: []}, routes)
        mock_route_delete.assert_called_once_with(route)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_del_ip_route_vlan(self, mock_route_delete):
        routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 32,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': 10}]}
        route = copy.deepcopy(routes[self.dev][0]['route'])

        linux_net.del_ip_route(routes, self.ip, 7, self.dev, vlan=10)

        self.assertEqual({self.dev: []}, routes)
        mock_route_delete.assert_called_once_with(route)

    @mock.patch('ovn_bgp_agent.privileged.linux_net.route_delete')
    def test_del_ip_route_mask(self, mock_route_delete):
        routes = {
            self.dev: [{'route': {'dst': self.ip,
                                  'dst_len': 30,
                                  'oif': mock.ANY,
                                  'proto': 3,
                                  'scope': 253,
                                  'table': 7},
                        'vlan': None}]}
        route = copy.deepcopy(routes[self.dev][0]['route'])

        linux_net.del_ip_route(routes, self.ip, 7, self.dev, mask=30)

        self.assertEqual({self.dev: []}, routes)
        mock_route_delete.assert_called_once_with(route)


class TestEnsureRoutingTableForBridge(test_base.TestCase):
    def setUp(self):
        super().setUp()
        self.ovn_routing_tables = {}
        self.bridge_name = "br-test"
        self.vrf_table = 4
        self.generated_number = 2

        self.testing_multiline_file_content = [
            "1 foo",
            "# commented line",
            "random garbage text",
            "3  another bridge",
            "2",
        ]

        self.m_ensure_rt_routes = mock.patch.object(
            linux_net, '_ensure_routing_table_routes').start()

        # The 'random' generator will always generate number 2 because the
        # range is 1-4 while 1 and 3 are used in the file and 4 is the vrf
        mock.patch.object(constants, 'ROUTING_TABLE_MIN', 1).start()
        mock.patch.object(constants, 'ROUTING_TABLE_MAX', 4).start()

    def _create_fake_file_content(self):
        return "\n".join(self.testing_multiline_file_content)

    def test_ensure_routing_table_for_bridge_table_missing(self):
        self._test_ensure_routing_table_for_bridge_table_missing()

    def _test_ensure_routing_table_for_bridge_table_missing(self):
        with mock.patch(
                'builtins.open',
                mock.mock_open(read_data=self._create_fake_file_content())):
            linux_net.ensure_routing_table_for_bridge(
                self.ovn_routing_tables, self.bridge_name, self.vrf_table)

        self.assertDictEqual(
            {self.bridge_name: self.generated_number}, self.ovn_routing_tables)

    def test_ensure_routing_table_for_bridge_table_present(self):
        present_bridge_value = 5
        self.testing_multiline_file_content.insert(
            2, "%d %s" % (present_bridge_value, self.bridge_name))

        with mock.patch(
                'builtins.open',
                mock.mock_open(read_data=self._create_fake_file_content())):
            linux_net.ensure_routing_table_for_bridge(
                self.ovn_routing_tables, self.bridge_name, self.vrf_table)

        self.assertDictEqual(
            {self.bridge_name: present_bridge_value}, self.ovn_routing_tables)

    def test_ensure_routing_table_for_bridge_table_vrf_not_generated(self):
        self.vrf_table = 2
        self.generated_number = 4
        self._test_ensure_routing_table_for_bridge_table_missing()

    def test_ensure_routing_table_for_bridge_tables_depleted(self):
        present_bridge_value = 2
        self.testing_multiline_file_content.insert(
            2, "%d %s" % (present_bridge_value, "foo"))

        with mock.patch(
                'builtins.open',
                mock.mock_open(read_data=self._create_fake_file_content())):
            self.assertRaises(
                SystemExit,
                linux_net.ensure_routing_table_for_bridge,
                self.ovn_routing_tables, self.bridge_name, self.vrf_table)

        self.assertDictEqual({}, self.ovn_routing_tables)



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
from unittest import mock

from oslo_concurrency import processutils

from ovn_bgp_agent.privileged import ovs_vsctl
from ovn_bgp_agent.tests import base as test_base

# Mock the privsep decorator and reload the module
mock.patch('ovn_bgp_agent.privileged.ovs_vsctl_cmd.entrypoint',
           lambda x: x).start()
importlib.reload(ovs_vsctl)


class FakeException(Exception):
    stderr = ''


class TestPrivilegedOvsVsctl(test_base.TestCase):

    def setUp(self):
        super(TestPrivilegedOvsVsctl, self).setUp()
        # Mock processutils.execute()
        self.mock_exc = mock.patch.object(processutils, 'execute').start()

    def test_ovs_vsctl(self):
        ovs_vsctl.ovs_vsctl(
            ['--if-exists', 'del-port', 'fake-port'])
        self.mock_exc.assert_called_once_with(
            'ovs-vsctl', '--if-exists', 'del-port', 'fake-port')

    def test_ovs_ofctl(self):
        ovs_vsctl.ovs_ofctl(
            ['dump-flows', 'dummy-br'])
        self.mock_exc.assert_called_once_with(
            'ovs-ofctl', 'dump-flows', 'dummy-br')

    def test_ovs_vsctl_timeout(self):
        ovs_vsctl.ovs_vsctl(
            ['--if-exists', 'del-port', 'fake-port'], timeout=10)
        self.mock_exc.assert_called_once_with(
            'ovs-vsctl', '--timeout=10', '--if-exists', 'del-port',
            'fake-port')

    def test_ovs_ofctl_timeout(self):
        ovs_vsctl.ovs_ofctl(
            ['dump-flows', 'dummy-br'], timeout=10)
        self.mock_exc.assert_called_once_with(
            'ovs-ofctl', '--timeout=10', 'dump-flows', 'dummy-br')

    def test_ovs_ofctl_fallback_OF_version(self):
        # fallback only applies to ovs-ofctl command
        self.mock_exc.side_effect = (
            processutils.ProcessExecutionError(), None)
        ovs_vsctl.ovs_ofctl(
            ['--strict', 'del-flows', 'br-ex', 'dummy-flow'])

        calls = [mock.call('ovs-ofctl', '--strict', 'del-flows',
                           'br-ex', 'dummy-flow'),
                 mock.call('ovs-ofctl', '--strict', 'del-flows',
                           'br-ex', 'dummy-flow', '-O', 'OpenFlow13')]
        self.mock_exc.assert_has_calls(calls)

    def test_ovs_vsctl_process_execution_error_no_fallback(self):
        # fallback does not apply to ovs-vsctl command
        self.mock_exc.side_effect = processutils.ProcessExecutionError()
        self.assertRaises(
            processutils.ProcessExecutionError, ovs_vsctl.ovs_vsctl,
            ['--if-exists', 'del-port', 'fake-port'])
        self.mock_exc.assert_called_once_with(
            'ovs-vsctl', '--if-exists', 'del-port', 'fake-port')

    def test_ovs_ofctl_exception(self):
        self.mock_exc.side_effect = FakeException()
        self.assertRaises(
            FakeException, ovs_vsctl.ovs_ofctl,
            ['add-flow', 'br-ex', 'dummy-flow'])
        self.mock_exc.assert_called_once_with(
            'ovs-ofctl', 'add-flow', 'br-ex', 'dummy-flow')

    def test_ovs_cmd_fallback_exception(self):
        self.mock_exc.side_effect = (
            processutils.ProcessExecutionError(), FakeException())
        self.assertRaises(
            FakeException, ovs_vsctl.ovs_ofctl,
            ['add-flow', 'br-ex', 'dummy-flow'])
        calls = [mock.call('ovs-ofctl', 'add-flow', 'br-ex', 'dummy-flow'),
                 mock.call('ovs-ofctl', 'add-flow', 'br-ex', 'dummy-flow',
                           '-O', 'OpenFlow13')]
        self.mock_exc.assert_has_calls(calls)

    def test_ovs_appctl(self):
        ovs_vsctl.ovs_appctl(
            ['fdb/add', 'br-fabric', 'br-fabric', '0', 'de:ad:be:ef:ca:fe'])
        self.mock_exc.assert_called_once_with(
            'ovs-appctl', 'fdb/add', 'br-fabric', 'br-fabric',
            '0', 'de:ad:be:ef:ca:fe')



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ipaddress
import random
import re
import sys

import netaddr
from neutron_lib import constants as n_const
from oslo_log import log as logging
import pyroute2
from pyroute2.netlink import exceptions as netlink_exceptions
import tenacity

from ovn_bgp_agent import constants
from ovn_bgp_agent import exceptions as agent_exc
import ovn_bgp_agent.privileged.linux_net
from ovn_bgp_agent.utils import common as common_utils

LOG = logging.getLogger(__name__)

RE_TABLE_ROW = re.compile(r"^(?P<table>[0-9]+)\s+(?P<bridge>\S+)")


def get_ip_version(ip):
    # IP network can consume both an IP address and a network with cidr
    # notation
    return netaddr.IPNetwork(ip).version


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_interfaces(filter_out=[]):
    with pyroute2.IPRoute() as ipr:
        return [iface.get_attr('IFLA_IFNAME') for iface in ipr.get_links()
                if iface.get_attr('IFLA_IFNAME') not in filter_out]


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_interface_index(nic):
    try:
        with pyroute2.IPRoute() as ipr:
            return ipr.link_lookup(ifname=nic)[0]
    except IndexError:
        raise agent_exc.NetworkInterfaceNotFound(device=nic)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_interface_address(nic):
    try:
        with pyroute2.IPRoute() as ipr:
            idx = ipr.link_lookup(ifname=nic)[0]
            return ipr.get_links(idx)[0].get_attr('IFLA_ADDRESS')
    except IndexError:
        raise agent_exc.NetworkInterfaceNotFound(device=nic)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_nic_info(nic):
    try:
        with pyroute2.IPRoute() as ipr:
            idx = ipr.link_lookup(ifname=nic)[0]
            nic_addr = ipr.get_addr(index=idx)[0]
            ip = '{}/{}'.format(
                nic_addr.get_attr('IFA_ADDRESS'),
                nic_addr.get('prefixlen'))
            mac = ipr.get_links(idx)[0].get_attr('IFLA_ADDRESS')
            return ip, mac
    except IndexError:
        raise agent_exc.NetworkInterfaceNotFound(device=nic)


def ensure_vrf(vrf_name, vrf_table):
    ovn_bgp_agent.privileged.linux_net.ensure_vrf(vrf_name, vrf_table)


def ensure_bridge(bridge_name):
    ovn_bgp_agent.privileged.linux_net.ensure_bridge(bridge_name)


def ensure_vxlan(vxlan_name, vni, local_ip, dstport):
    ovn_bgp_agent.privileged.linux_net.ensure_vxlan(vxlan_name, vni, local_ip,
                                                    dstport)


def ensure_veth(veth_name, veth_peer):
    ovn_bgp_agent.privileged.linux_net.ensure_veth(veth_name, veth_peer)


def set_master_for_device(device, master):
    ovn_bgp_agent.privileged.linux_net.set_master_for_device(device, master)


def ensure_dummy_device(device):
    ovn_bgp_agent.privileged.linux_net.ensure_dummy_device(device)


def ensure_ovn_device(ovn_ifname, vrf_name):
    ensure_dummy_device(ovn_ifname)
    set_master_for_device(ovn_ifname, vrf_name)


def delete_device(device):
    ovn_bgp_agent.privileged.linux_net.delete_device(device)


def ensure_arp_ndp_enabled_for_bridge(bridge, offset, vlan_tag=None):
    ipv4 = "%s%d.%s" % (
        constants.ARP_IPV4_PREFIX, offset / constants.IPV4_OCTET_RANGE,
        offset % constants.IPV4_OCTET_RANGE)
    ipv6 = "%s%x" % (constants.NDP_IPV6_PREFIX, offset)

    for ip in (ipv4, ipv6):
        try:
            ovn_bgp_agent.privileged.linux_net.add_ip_to_dev(ip, bridge)
        except agent_exc.IpAddressAlreadyExists:
            LOG.debug("IP %s already added on bridge %s", ip, bridge)
        except KeyError as e:
            if "object exists" not in str(e):
                LOG.error("Unable to add IP on bridge %s to enable arp/ndp. "
                          "Exception: %s", bridge, e)
                raise

    # also enable the arp/ndp on the bridge in case there are flat networks
    enable_proxy_arp(bridge)
    enable_proxy_ndp(bridge)


def ensure_anycast_mac_for_interface(intf, offset):
    # Make pointer to module, to shorten amount of chars to call module.
    priv = ovn_bgp_agent.privileged.linux_net

    # The intf acting as L3 GW (needs to have same mac address everywhere)
    # So generate the mac address based on the given offset and add the
    # configured MAC_LLADR_OFFSET to it.
    mac_int = int(constants.MAC_LLADDR_OFFSET.replace(':', ''), 16)
    ll = ('%x' % int(mac_int + offset)).zfill(12)
    lladdr = ":".join([ll[i:i + 2] for i in range(0, len(ll), 2)])

    # Check what the mac address currently is.
    dev = priv.get_link_device(intf)
    curr_lladdr = priv.get_attr(dev, 'IFLA_ADDRESS')

    if lladdr != curr_lladdr:
        LOG.info("Updating mac address for intf %s to address %s",
                 intf, lladdr)
        priv.set_link_attribute(intf, address=lladdr)

        # Also update the 'scope link' address on the interface.
        ll = lladdr.replace(':', '')
        ll_ip_address = ipaddress.IPv6Address(
            f'fe80::{ll[0:4]}:{ll[4:8]}:{ll[8:12]}')  # noqa: E231
        ll_net = ipaddress.IPv6Network('fe80::/10')

        # Fetch all ipv6 addresses and check if we already configured the
        # link-local address.
        addresses = priv.get_ip_addresses(
            index=dev['index'], family=constants.AF_INET6)
        for addr in addresses:
            ip_addr = ipaddress.IPv6Address(
                priv.get_attr(addr, 'IFA_ADDRESS'))

            if (addr['scope'] != 0 and
                    ip_addr in ll_net and
                    ip_addr != ll_ip_address):

                LOG.info('Update scope link address on intf %s from %s to %s',
                         intf, ip_addr, ll_ip_address)

                # Delete the old link local ip address from the interface
                priv.delete_ip_address(ip_addr.compressed, intf,
                                       prefixlen=addr['prefixlen'],
                                       scope=addr['scope'])

                # Attach our anycast link local address to the interface
                priv.add_ip_address(ll_ip_address.compressed, intf,
                                    prefixlen=addr['prefixlen'],
                                    scope=addr['scope'])


def disable_learning_vxlan_intf(intf):
    '''ip link set vni200 type bridge_slave neigh_suppress on learning off'''
    ovn_bgp_agent.privileged.linux_net.set_brport_attribute(
        intf, neigh_suppress=True, learning=False
    )


def ensure_routing_table_for_bridge(ovn_routing_tables, bridge, vrf_table):
    # check a routing table with the bridge name exists on
    # /etc/iproute2/rt_tables
    found_tables = {vrf_table}

    with open(constants.ROUTING_TABLES_FILE, 'r') as rt_file:
        for line in rt_file.readlines():
            match = RE_TABLE_ROW.match(line)
            if match:
                if match.group('bridge') == bridge:
                    # We don't need to catch exception for TypeError because
                    # the regular expression matches only integers
                    ovn_routing_tables[match.group('bridge')] = int(
                        match.group('table'))
                    LOG.debug("Found routing table for %s with: %s",
                              bridge, match.group('table'))
                    break
                else:
                    found_tables.add(int(match.group('table')))
        else:
            LOG.debug("Routing table for bridge %s not configured at ", bridge)
            try:
                routing_table_range = set(
                    range(constants.ROUTING_TABLE_MIN,
                          constants.ROUTING_TABLE_MAX + 1))
                table_number = random.choice(
                    list(routing_table_range - found_tables))
            except IndexError:
                LOG.error("No more routing tables available for bridge %s "
                          "at %s", constants.ROUTING_TABLES_FILE, bridge)
                sys.exit(1)
            ovn_bgp_agent.privileged.linux_net.create_routing_table_for_bridge(
                table_number, bridge)
            ovn_routing_tables[bridge] = int(table_number)
            LOG.debug("Added routing table for %s with number: %s",
                      bridge, table_number)

    return _ensure_routing_table_routes(ovn_routing_tables, bridge)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def _ensure_routing_table_routes(ovn_routing_tables, bridge):
    # add default route on that table if it does not exist
    extra_routes = []
    bridge_idx = get_interface_index(bridge)

    with pyroute2.IPRoute() as ip:
        table_route_dsts = {
            (r.get_attr('RTA_DST'), r['dst_len'])
            for r in ip.get_routes(table=ovn_routing_tables[bridge])
        }

        if not table_route_dsts:
            r1 = {'dst': 'default', 'oif': bridge_idx,
                  'table': ovn_routing_tables[bridge], 'scope': 253,
                  'proto': 3}
            ovn_bgp_agent.privileged.linux_net.route_create(r1)

            r2 = {'dst': 'default', 'oif': bridge_idx,
                  'table': ovn_routing_tables[bridge],
                  'family': constants.AF_INET6,
                  'proto': 3}
            ovn_bgp_agent.privileged.linux_net.route_create(r2)
        else:
            route_missing = True
            route6_missing = True
            for (dst, dst_len) in table_route_dsts:
                if not dst:  # default route
                    try:
                        route = [
                            r for r in ip.get_routes(
                                table=ovn_routing_tables[bridge],
                                family=constants.AF_INET)
                            if not r.get_attr('RTA_DST')][0]
                        if bridge_idx == route.get_attr('RTA_OIF'):
                            route_missing = False
                        else:
                            extra_routes.append(route)
                    except IndexError:
                        pass  # no ipv4 default rule
                    try:
                        route_6 = [
                            r for r in ip.get_routes(
                                table=ovn_routing_tables[bridge],
                                family=constants.AF_INET6)
                            if not r.get_attr('RTA_DST')][0]
                        if bridge_idx == route_6.get_attr('RTA_OIF'):
                            route6_missing = False
                        else:
                            extra_routes.append(route_6)
                    except IndexError:
                        pass  # no ipv6 default rule
                else:
                    if get_ip_version(dst) == constants.IP_VERSION_6:
                        extra_routes.append(
                            ip.get_routes(
                                table=ovn_routing_tables[bridge],
                                dst=dst,
                                dst_len=dst_len,
                                family=constants.AF_INET6)[0])
                    else:
                        extra_routes.append(
                            ip.get_routes(
                                table=ovn_routing_tables[bridge],
                                dst=dst,
                                dst_len=dst_len,
                                family=constants.AF_INET)[0])

            if route_missing:
                r = {'dst': 'default', 'oif': bridge_idx,
                     'table': ovn_routing_tables[bridge], 'scope': 253,
                     'proto': 3}
                ovn_bgp_agent.privileged.linux_net.route_create(r)
            if route6_missing:
                r = {'dst': 'default', 'oif': bridge_idx,
                     'table': ovn_routing_tables[bridge],
                     'family': constants.AF_INET6,
                     'proto': 3}
                ovn_bgp_agent.privileged.linux_net.route_create(r)
    return extra_routes


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_extra_routing_table_for_bridge(ovn_routing_tables, bridge):
    extra_routes = []
    bridge_idx = get_interface_index(bridge)
    with pyroute2.IPRoute() as ip:
        table_route_dsts = {
            (r.get_attr('RTA_DST'), r['dst_len'])
            for r in ip.get_routes(table=ovn_routing_tables[bridge])
        }

        if not table_route_dsts:
            return extra_routes

        for (dst, dst_len) in table_route_dsts:
            if not dst:  # default route
                try:
                    route = [
                        r for r in ip.get_routes(
                            table=ovn_routing_tables[bridge],
                            family=constants.AF_INET)
                        if not r.get_attr('RTA_DST')][0]
                    if bridge_idx != route.get_attr('RTA_OIF'):
                        extra_routes.append(route)
                except IndexError:
                    pass  # no IPv4 default rule
                try:
                    route_6 = [
                        r for r in ip.get_routes(
                            table=ovn_routing_tables[bridge],
                            family=constants.AF_INET6)
                        if not r.get_attr('RTA_DST')][0]
                    if bridge_idx != route_6.get_attr('RTA_OIF'):
                        extra_routes.append(route_6)
                except IndexError:
                    pass  # no IPv6 default rule
            else:
                if get_ip_version(dst) == constants.IP_VERSION_6:
                    extra_routes.append(
                        ip.get_routes(
                            table=ovn_routing_tables[bridge],
                            dst=dst,
                            dst_len=dst_len,
                            family=constants.AF_INET6)[0])
                else:
                    extra_routes.append(
                        ip.get_routes(
                            table=ovn_routing_tables[bridge],
                            dst=dst,
                            dst_len=dst_len,
                            family=constants.AF_INET)[0])
    return extra_routes


def ensure_vlan_device_for_network(bridge, vlan_tag):
    ovn_bgp_agent.privileged.linux_net.ensure_vlan_device_for_network(bridge,
                                                                      vlan_tag)
    device = "{}/{}".format(
        bridge[:constants.OVN_VLAN_DEVICE_MAX_LENGTH], vlan_tag)
    enable_proxy_arp(device)
    enable_proxy_ndp(device)


def delete_vlan_device_for_network(bridge, vlan_tag):
    vlan_device_name = '{}.{}'.format(
        bridge[:constants.OVN_VLAN_DEVICE_MAX_LENGTH], vlan_tag)
    delete_device(vlan_device_name)


def get_bridge_vlans(bridge):
    return ovn_bgp_agent.privileged.linux_net.get_bridge_vlans(bridge)


def enable_proxy_ndp(device):
    flag = f"net.ipv6.conf.{device[:n_const.DEVICE_NAME_MAX_LEN]}.proxy_ndp"
    ovn_bgp_agent.privileged.linux_net.set_kernel_flag(flag, 1)


def enable_proxy_arp(device):
    flag = f"net.ipv4.conf.{device[:n_const.DEVICE_NAME_MAX_LEN]}.proxy_arp"
    ovn_bgp_agent.privileged.linux_net.set_kernel_flag(flag, 1)


def enable_routing_for_interfaces(*interfaces):
    # Configure sysctl
    keys = [
        ('net.ipv4.ip_forward', 1),
        ('net.ipv4.conf.all.forwarding', 1),
        ('net.ipv6.conf.all.forwarding', 1),
    ]
    for intf in interfaces:
        intf_key = intf.replace('.', '/')
        keys.append((f'net.ipv4.conf.{intf_key}.forwarding', 1))
        keys.append((f'net.ipv6.conf.{intf_key}.forwarding', 1))

    for k, v in keys:
        LOG.debug('Configure sysctl %s=%s', k, v)
        ovn_bgp_agent.privileged.linux_net.set_kernel_flag(k, v)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_exposed_ips(nic):
    nic_idx = get_interface_index(nic)
    try:
        with pyroute2.IPRoute() as ipr:
            return [ip.get_attr('IFA_ADDRESS')
                    for ip in ipr.get_addr(index=nic_idx)
                    if ip['prefixlen'] in (32, 128)]
    except pyroute2.netlink.exceptions.NetlinkError:
        # Nic does not exist
        LOG.debug("NIC %s does not yet exist, so it does not have exposed IPs",
                  nic)
        return []


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_nic_ip(nic, prefixlen_filter=None):
    nic_idx = get_interface_index(nic)
    with pyroute2.IPRoute() as ipr:
        if prefixlen_filter:
            return [
                ip.get_attr('IFA_ADDRESS')
                for ip in ipr.get_addr(index=nic_idx,
                                       prefixlen=prefixlen_filter)
            ]
        else:
            return [
                ip.get_attr('IFA_ADDRESS')
                for ip in ipr.get_addr(index=nic_idx)
            ]


def get_exposed_ips_on_network(nic, network):
    exposed_ips = get_exposed_ips(nic)
    return [ip for ip in exposed_ips if ipaddress.ip_address(ip) in network]


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_exposed_routes_on_network(table_ids, network):
    with pyroute2.NDB() as ndb:
        # NOTE: skip bgp routes (proto 186)
        return [
            r
            for r in ndb.routes.dump()
            if r.table in table_ids and
            r.dst != "" and
            r.gateway is not None and
            r.proto != 186 and
            ipaddress.ip_address(r.gateway) in network
        ]


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_ovn_ip_rules(routing_tables):
    ovn_ip_rules = {}
    with pyroute2.IPRoute() as ipr:
        rules_info = [
            (rule.get_attr('FRA_TABLE'),
             "{}/{}".format(rule.get_attr('FRA_DST'), rule['dst_len']),
             rule['family'])
            for rule in (
                ipr.get_rules(family=constants.AF_INET) +
                ipr.get_rules(family=constants.AF_INET6))
            if rule.get_attr('FRA_TABLE') in routing_tables
        ]
        for table, dst, family in rules_info:
            ovn_ip_rules[dst] = {'table': table, 'family': family}
    return ovn_ip_rules


def delete_exposed_ips(ips, nic):
    ovn_bgp_agent.privileged.linux_net.delete_exposed_ips(ips, nic)


def delete_ip_rules(ip_rules):
    ovn_bgp_agent.privileged.linux_net.delete_ip_rules(ip_rules)


def delete_bridge_ip_routes(routing_tables, routing_tables_routes,
                            extra_routes):
    for device, routes_info in routing_tables_routes.items():
        if not extra_routes.get(device):
            continue
        for route_info in routes_info:
            oif = get_interface_index(device)
            if route_info['vlan']:
                vlan_device_name = '{}.{}'.format(
                    device[:constants.OVN_VLAN_DEVICE_MAX_LENGTH],
                    route_info['vlan'])
                oif = get_interface_index(vlan_device_name)
            if 'gateway' in route_info['route'].keys():  # subnet route
                possible_matchings = [
                    r for r in extra_routes[device]
                    if (r.get_attr('RTA_DST') == route_info['route']['dst'] and
                        r['dst_len'] == route_info['route']['dst_len'] and
                        r.get_attr('RTA_GATEWAY') == route_info['route'][
                            'gateway'])]
            else:  # cr-lrp
                possible_matchings = [
                    r for r in extra_routes[device]
                    if (r.get_attr('RTA_DST') == route_info['route']['dst'] and
                        r['dst_len'] == route_info['route']['dst_len'] and
                        r.get_attr('RTA_OIF') == oif)]
            for r in possible_matchings:
                extra_routes[device].remove(r)

    for bridge, routes in extra_routes.items():
        for route in routes:
            r_info = {'dst': route.get_attr('RTA_DST'),
                      'dst_len': route['dst_len'],
                      'family': route['family'],
                      'oif': route.get_attr('RTA_OIF'),
                      'table': routing_tables[bridge]}
            if route.get_attr('RTA_GATEWAY'):
                r_info['gateway'] = route.get_attr('RTA_GATEWAY')
            ovn_bgp_agent.privileged.linux_net.route_delete(r_info)


def delete_routes_from_table(table):
    table_routes = _get_table_routes(table)
    delete_ip_routes(table_routes)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def _get_table_routes(table):
    with pyroute2.IPRoute() as ipr:
        return [
            r for r in ipr.get_routes(table=table)
            if r['scope'] != 254 and r['proto'] != 186
        ]


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def get_routes_on_tables(table_ids):
    routes = []
    with pyroute2.IPRoute() as ipr:
        for table_id in table_ids:
            table_routes = [
                r for r in ipr.get_routes(table=table_id)
                if r.get_attr('RTA_DST') and r['proto'] != 186
            ]
            routes.extend(table_routes)
    return routes


def delete_ip_routes(routes):
    for route in routes:
        r_info = {'dst': route.get('dst'),
                  'dst_len': route['dst_len'],
                  'family': route['family'],
                  'oif': route.get('oif'),
                  'gateway': route.get('gateway'),
                  'table': route['table']}
        ovn_bgp_agent.privileged.linux_net.route_delete(r_info)


def add_ndp_proxy(ip, dev, vlan=None):
    ovn_bgp_agent.privileged.linux_net.add_ndp_proxy(ip, dev, vlan)


def del_ndp_proxy(ip, dev, vlan=None):
    ovn_bgp_agent.privileged.linux_net.del_ndp_proxy(ip, dev, vlan)


def add_ips_to_dev(nic, ips, clear_local_route_at_table=False):
    already_added_ips = []
    for ip in ips:
        try:
            ovn_bgp_agent.privileged.linux_net.add_ip_to_dev(ip, nic)
        except agent_exc.IpAddressAlreadyExists:
            already_added_ips.append(ip)

    if clear_local_route_at_table:
        for ip in ips:
            if ip in already_added_ips:
                continue
            oif = get_interface_index(nic)
            route = {'table': clear_local_route_at_table,
                     'proto': 2,
                     'scope': 254,
                     'dst': ip,
                     'oif': oif}
            ovn_bgp_agent.privileged.linux_net.route_delete(route)


def del_ips_from_dev(nic, ips):
    for ip in ips:
        ovn_bgp_agent.privileged.linux_net.del_ip_from_dev(ip, nic)


def create_rule_from_ip(ip, table):
    try:
        ip_network = netaddr.IPNetwork(ip)
    except (netaddr.AddrFormatError, ValueError):
        raise agent_exc.InvalidPortIP(ip=ip)

    return {
        'dst': str(ip_network.ip),
        'table': table,
        'dst_len': ip_network.prefixlen,
        'family': common_utils.IP_VERSION_FAMILY_MAP[ip_network.version],
    }


def add_ip_rule(ip, table, dev=None, lladdr=None):
    rule = create_rule_from_ip(ip, table)

    ovn_bgp_agent.privileged.linux_net.rule_create(rule)

    if lladdr:
        add_ip_nei(ip, lladdr, dev)


def add_ip_nei(ip, lladdr, dev):
    """Add ip neighbor permament entry

    param ip: IP of the neighbor to add an entry for
    param lladdr: link layer address of the neighbor to associate to that IP
    param dev: the interface to which the neighbor is attached
    """
    ovn_bgp_agent.privileged.linux_net.add_ip_nei(ip, lladdr, dev)


def del_ip_rule(ip, table, dev=None, lladdr=None):
    rule = create_rule_from_ip(ip, table)

    ovn_bgp_agent.privileged.linux_net.rule_delete(rule)

    # TODO(jayjahns): The lladdr in all delete operations is blank, so
    # we need to investigate how to ensure that exists in order to
    # add the if comparison.
    del_ip_nei(ip, lladdr, dev)


def del_ip_nei(ip, lladdr, dev):
    """Del ip neighbor permament entry

    param ip: IP of the neighbor to delete the entry
    param lladdr: link layer address of the neighbor to disassociate
    param dev: the interface to which the neighbor is attached
    """
    ovn_bgp_agent.privileged.linux_net.del_ip_nei(ip, lladdr, dev)


def add_unreachable_route(vrf_name):
    ovn_bgp_agent.privileged.linux_net.add_unreachable_route(vrf_name)


@tenacity.retry(
    retry=tenacity.retry_if_exception_type(
        netlink_exceptions.NetlinkDumpInterrupted),
    wait=tenacity.wait_exponential(multiplier=0.02, max=1),
    stop=tenacity.stop_after_delay(8),
    reraise=True)
def add_ip_route(ovn_routing_tables_routes, ip_address, route_table, dev,
                 vlan=None, mask=None, via=None):
    net_ip = ip_address
    if not mask:  # default /32 or /128
        if get_ip_version(ip_address) == constants.IP_VERSION_6:
            mask = 128
        else:
            mask = 32
    else:
        ip = '{}/{}'.format(ip_address, mask)
        if get_ip_version(ip_address) == constants.IP_VERSION_6:
            net_ip = '{}'.format(ipaddress.IPv6Network(
                ip, strict=False).network_address)
        else:
            net_ip = '{}'.format(ipaddress.IPv4Network(
                ip, strict=False).network_address)

    if vlan:
        oif_name = '{}.{}'.format(
            dev[:constants.OVN_VLAN_DEVICE_MAX_LENGTH], vlan)
        try:
            oif = get_interface_index(oif_name)
        except agent_exc.NetworkInterfaceNotFound:
            # Most provider network was recently created an
            # there has not been a sync since then, therefore
            # the vlan device has not yet been created
            # Trying to create the device and retrying
            ensure_vlan_device_for_network(dev, vlan)
            oif = get_interface_index(oif_name)
    else:
        oif = get_interface_index(dev)

    route = {'dst': net_ip, 'dst_len': int(mask), 'oif': oif,
             'table': int(route_table), 'proto': 3}
    if via:
        route['gateway'] = via
        route['scope'] = 0
    else:
        route['scope'] = 253
    if get_ip_version(net_ip) == constants.IP_VERSION_6:
        route['family'] = constants.AF_INET6
        del route['scope']

    with pyroute2.IPRoute() as ipr:
        if not ipr.route('show', **route):
            LOG.debug("Creating route at table %s: %s", route_table, route)
            ovn_bgp_agent.privileged.linux_net.route_create(route)
            LOG.debug("Route created at table %s: %s", route_table, route)
        else:
            LOG.debug("Route already existing: %s", route)
    route_info = {'vlan': vlan, 'route': route}
    ovn_routing_tables_routes.setdefault(dev, []).append(route_info)


def del_ip_route(ovn_routing_tables_routes, ip_address, route_table, dev,
                 vlan=None, mask=None, via=None):
    net_ip = ip_address
    if not mask:  # default /32 or /128
        if get_ip_version(ip_address) == constants.IP_VERSION_6:
            mask = 128
        else:
            mask = 32
    else:
        ip = '{}/{}'.format(ip_address, mask)
        if get_ip_version(ip_address) == constants.IP_VERSION_6:
            net_ip = '{}'.format(ipaddress.IPv6Network(
                ip, strict=False).network_address)
        else:
            net_ip = '{}'.format(ipaddress.IPv4Network(
                ip, strict=False).network_address)

    try:
        if vlan:
            oif_name = '{}.{}'.format(
                dev[:constants.OVN_VLAN_DEVICE_MAX_LENGTH], vlan)
            oif = get_interface_index(oif_name)
        else:
            oif = get_interface_index(dev)
    except agent_exc.NetworkInterfaceNotFound:
        LOG.debug("Device %s does not exists, so the associated "
                  "routes should have been automatically deleted.", dev)
        ovn_routing_tables_routes.pop(dev, None)
        return

    route = {'dst': net_ip, 'dst_len': int(mask), 'oif': oif,
             'table': int(route_table), 'proto': 3}
    if via:
        route['gateway'] = via
        route['scope'] = 0
    else:
        route['scope'] = 253
    if get_ip_version(net_ip) == constants.IP_VERSION_6:
        route['family'] = constants.AF_INET6
        del route['scope']

    LOG.debug("Deleting route at table %s: %s", route_table, route)
    ovn_bgp_agent.privileged.linux_net.route_delete(route)
    LOG.debug("Route deleted at table %s: %s", route_table, route)
    route_info = {'vlan': vlan, 'route': route}
    if route_info in ovn_routing_tables_routes.get(dev, []):
        ovn_routing_tables_routes[dev].remove(route_info)


def set_device_status(device, status, ndb=None):
    ovn_bgp_agent.privileged.linux_net.set_device_state(
        device, status, ndb=ndb)
